"""
YAML path Search segment terms.

Copyright 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
from yamlpath.enums import PathSearchMethods


class SearchTerms:
    """YAML path Search segment terms."""

    def __init__(
        self, inverted: bool, method: PathSearchMethods, attribute: str,
        term: str
    ) -> None:
        """
        Instantiate a Search Term.

        Parameters:
        1. inverted (bool) true = invert the search results; false, otherwise
        2. method (PathSearchMethods) the method of search
        3. attribute (str) the attribute to search
        4. term (str) the term to search for within attribute
        """
        self._inverted: bool = inverted
        self._method: PathSearchMethods = method
        self._attribute: str = attribute
        self._term: str = term

    # While this works in Python 3.7.3, it does not work in Python 3.6.3.  In
    # the older Python, this code creates a cyclic ImportError.  Because this
    # works in newer Pythons, I'm leaving this code here but commented-out
    # should I ever decide to stop supporting older Pythons.  Until then, I'm
    # moving this code to a neutral helper function that'll import both
    # SearchTerms and PathAttributes, thus neutering this nonsense.
    # @classmethod
    # def from_path_segment_attrs(
    #         cls: Type,
    #         rhs: "pathattributes.PathAttributes") -> "SearchTerms":
    #     """
    #     Generates a new SearchTerms instance by copying SearchTerms
    #     attributes from a YAML Path segment's attributes.
    #     """
    #     if isinstance(rhs, SearchTerms):
    #         newinst: SearchTerms = cls(
    #             rhs.inverted, rhs.method, rhs.attribute, rhs.term
    #         )
    #         return newinst
    #     raise AttributeError

    def __str__(self) -> str:
        """Get a String representation of this Search Term."""
        if self.method == PathSearchMethods.REGEX:
            safe_term = "/{}/".format(self.term.replace("/", r"\/"))
        else:
            # Replace unescaped spaces with escaped spaces
            safe_term = r"\ ".join(
                list(map(
                    lambda ele: ele.replace(" ", r"\ ")
                    , self.term.split(r"\ ")
                ))
            )

        return (
            "["
            + str(self.attribute)
            + ("!" if self.inverted else "")
            + str(self.method)
            + safe_term
            + "]"
        )

    @property
    def inverted(self) -> bool:
        """
        Access the inversion flag for this Search.

        This indicates whether the results are to be inverted.
        """
        return self._inverted

    @property
    def method(self) -> PathSearchMethods:
        """
        Access the search method.

        This indicates what kind of search is to be performed.
        """
        return self._method

    @property
    def attribute(self) -> str:
        """
        Accessor for the attribute being searched.

        This is the "haystack" and may reference a particular dictionary key,
        all values of a dictionary, or the elements of a list.
        """
        return self._attribute

    @property
    def term(self) -> str:
        """
        Accessor for the search term.

        This is the "needle" to search for within the attribute ("haystack").
        """
        return self._term



"""
Validate YAML/JSON/Compatible data.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
import sys
import argparse

from yamlpath import __version__ as YAMLPATH_VERSION
from yamlpath.common import Parsers
from yamlpath.wrappers import ConsolePrinter

class LogErrorCap:
    """Capture only ERROR messages as a fake ConsolePrinter."""

    def __init__(self):
        """Initialize this class instance."""
        self.lines = []
    def info(self, message):
        """Discard INFO messages."""
    def verbose(self, message):
        """Discard verbose INFO messages."""
    def warning(self, message):
        """Discard WARNING messages."""
    # pylint: disable=unused-argument
    def error(self, message, *args):
        """Capture ERROR messages."""
        self.lines.append(message)
    # pylint: disable=unused-argument
    def critical(self, message, *args):
        """Discard critical ERROR messages."""
    def debug(self, message, **kwargs):
        """Discard DEBUG messages."""

def processcli():
    """Process command-line arguments."""
    parser = argparse.ArgumentParser(
        description="Validate YAML, JSON, and compatible files.",
        epilog=(
            "Except when suppressing all report output with --quiet|-q,"
            " validation issues are printed to STDOUT (not STDERR).  Further,"
            " the exit-state will report 0 when there are no issues, 1 when"
            " there is an issue with the supplied command-line arguments, or 2"
            " when validation has failed for any document.  To report issues"
            " with this tool or to request enhancements, please visit"
            " https://github.com/wwkimball/yamlpath/issues.")
    )
    parser.add_argument("-V", "--version", action="version",
                        version="%(prog)s " + YAMLPATH_VERSION)

    parser.add_argument(
        "-S", "--nostdin", action="store_true",
        help=(
            "Do not implicitly read from STDIN, even when there are\n"
            "no - pseudo-files in YAML_FILEs with a non-TTY session"))

    noise_group = parser.add_mutually_exclusive_group()
    noise_group.add_argument(
        "-d", "--debug",
        action="store_true",
        help="output debugging details")
    noise_group.add_argument(
        "-v", "--verbose",
        action="store_true",
        help="increase output verbosity (show valid documents)")
    noise_group.add_argument(
        "-q", "--quiet",
        action="store_true",
        help="suppress all output except system errors")

    parser.add_argument("yaml_files", metavar="YAML_FILE", nargs="*",
                        help="one or more single- or multi-document"
                        " YAML/JSON/compatible files to validate; omit or use"
                        " - to read from STDIN")

    return parser.parse_args()

def validateargs(args, log):
    """Validate command-line arguments."""
    has_errors = False

    # There must be at least one input file or stream
    input_file_count = len(args.yaml_files)
    if (input_file_count == 0 and (
            sys.stdin.isatty()
            or args.nostdin)
    ):
        has_errors = True
        log.error(
            "There must be at least one YAML_FILE or STDIN document.")

    # There can be only one -
    pseudofile_count = 0
    for infile in args.yaml_files:
        if infile.strip() == '-':
            pseudofile_count += 1
    if pseudofile_count > 1:
        has_errors = True
        log.error("Only one YAML_FILE may be the - pseudo-file.")

    if has_errors:
        sys.exit(1)

def process_file(log, yaml, yaml_file):
    """Process a (potentially multi-doc) YAML file."""
    logcap = LogErrorCap()
    subdoc_index = 0
    exit_state = 0
    file_name = "STDIN" if yaml_file.strip() == "-" else yaml_file
    for (_, doc_loaded) in Parsers.get_yaml_multidoc_data(
        yaml, logcap, yaml_file
    ):
        if doc_loaded:
            log.verbose("{}/{} is valid.".format(file_name, subdoc_index))
        else:
            # An error message has been captured
            exit_state = 2
            log.info(
                "{}/{} is invalid due to:".format(file_name, subdoc_index))
            for line in logcap.lines:
                log.info("  * {}".format(line))
        logcap.lines.clear()
        subdoc_index += 1

    return exit_state

def main():
    """Main code."""
    # Process any command-line arguments
    args = processcli()
    log = ConsolePrinter(args)
    validateargs(args, log)
    exit_state = 0
    consumed_stdin = False
    yaml = Parsers.get_yaml_editor()

    for yaml_file in args.yaml_files:
        if yaml_file.strip() == '-':
            consumed_stdin = True

        log.debug(
            "yaml_merge::main:  Processing file, {}".format(
                "STDIN" if yaml_file.strip() == "-" else yaml_file))

        proc_state = process_file(log, yaml, yaml_file)

        if proc_state != 0:
            exit_state = proc_state

    # Check for a waiting STDIN document
    if (exit_state == 0
        and not consumed_stdin
        and not args.nostdin
        and not sys.stdin.isatty()
    ):
        exit_state = process_file(log, yaml, "-")

    sys.exit(exit_state)

if __name__ == "__main__":
    main()  # pragma: no cover



"""
Implements a reusable console print facility for simple command-line scripts.

Other implementations can easily wrap Python's standard logger/warning modules,
but this one does not because those are overkill for *simple* STDOUT/STDERR
printing (that must support squelching).

Requires an object on init which has the following properties:
  quiet:  <Boolean> suppresses all output except ConsolePrinter::error() and
          ::critical().
  verbose:  <Boolean> allows output from ConsolePrinter::verbose().
  debug:  <Boolean> allows output from ConsolePrinter::debug().

Copyright 2018, 2019, 2020, 2021 William W. Kimball, Jr. MBA MSIS
"""
import sys
from collections import deque
from typing import Any, Deque, Dict, Generator, List, Set, Tuple, Union

from ruamel.yaml.comments import (
    CommentedBase,
    CommentedMap,
    CommentedSet,
    TaggedScalar
)
# pylint: disable=wrong-import-position,ungrouped-imports
from ruamel.yaml import version_info as ryversion
if ryversion < (0, 17, 22):                   # pragma: no cover
    from yamlpath.patches.timestamp import (
        AnchoredTimeStamp,
        AnchoredDate,
    )  # type: ignore
else:                                         # pragma: no cover
    # Temporarily fool MYPY into resolving the future-case imports
    from ruamel.yaml.timestamp import TimeStamp as AnchoredTimeStamp
    AnchoredDate = AnchoredTimeStamp
    #from ruamel.yaml.timestamp import AnchoredTimeStamp
    # From whence shall come AnchoredDate?

from yamlpath.wrappers.nodecoords import NodeCoords
# pylint: enable=wrong-import-position,ungrouped-imports


class ConsolePrinter:
    """
    Generally-useful console messager.

    Writes INFO, VERBOSE, WARN, and DEBUG messages to STDOUT as well as ERROR
    messages to STDERR with multi-lne formatting.
    """

    def __init__(self, args):
        """
        Instantiate a ConsolePrinter.

        Positional Parameters:
        1. args (object) An object representing log level settings with these
           properties:
            - debug (Boolean) true = write debugging informational messages
            - verbose (Boolean) true = write verbose informational messages
            - quiet (Boolean) true = write only error messages

        Returns:  N/A

        Raises:  N/A
        """
        self.args = args

    def info(self, message):
        """
        Write an informational message to STDOUT unless quiet mode is active.

        Positional Parameters:
        1. message (str) The message to print

        Returns:  N/A

        Raises:  N/A
        """
        if not self.args.quiet:
            print(message)

    def verbose(self, message):
        """
        Write a verbose message to STDOUT.

        Writes only when verbose mode is active unless quiet mode is active.

        Positional Parameters:
        1. message (str) The message to print

        Returns:  N/A

        Raises:  N/A
        """
        if not self.args.quiet and (self.args.verbose or self.args.debug):
            print(message)

    def warning(self, message):
        """
        Write a warning message to STDOUT unless quiet mode is active.

        Positional Parameters:
        1. message (str) The message to print

        Returns:  N/A

        Raises:  N/A
        """
        if not self.args.quiet:
            print("WARNING:  " + str(message).replace("\n", "\nWARNING:  "))

    def error(self, message, exit_code=None):
        """
        Write a recoverable error message to STDERR.

        Optionally terminates the program, exiting with a specific error code.

        Positional Parameters:
        1. message (str) The message to print
        2. exit_code (int) The exit code to terminate the program with;
           default=None

        Returns:  N/A

        Raises:  N/A
        """
        print(
            "ERROR:  " + str(message).replace("\n", "\nERROR:  "),
            file=sys.stderr
        )
        print("Please try --help for more information.")
        sys.stdout.flush()

        # Optionally terminate program execution with a specified exit code
        if exit_code is not None:
            self.debug("Terminating with exit code, {}.".format(exit_code))
            sys.exit(exit_code)

    def critical(self, message, exit_code=1):
        """
        Write a critical, nonrecoverable failure message to STDERR and abend.

        Terminates the program, exiting with a specific error code.

        Positional Parameters:
        1. message (str) The message to print
        2. exit_code (int) The exit code to terminate the program with;
           default=1

        Returns:  N/A

        Raises:  N/A
        """
        print(
            "CRITICAL:  " + str(message).replace("\n", "\nCRITICAL:  "),
            file=sys.stderr
        )
        sys.stdout.flush()

        # Terminate program execution with a specified exit code
        self.debug("Terminating with exit code, {}.".format(exit_code))
        sys.exit(exit_code)

    def debug(self, message, **kwargs):
        """
        Write a debug message to STDOUT unless quiet mode is active.

        Dumps all key-value pairs of a dictionary or all elements of a list,
        when the message is either.

        Positional Parameters:
        1. message (str) The message to print

        Keyword Arguments:
        * data (Any) Data to recursively add to the DEBUG message
        * header (str) Line printed before the body of the DEBUG message
        * footer (str) Line printed after the body of the DEBUG message
        * prefix (str) String prefixed to every DEBUG line comprising the
          entirety of the DEBUG message, including any optional data
        * data_header (str) Line printed before the optional data, if any
        * data_footer (str) Line printed after the optional data, if any

        Returns:  N/A

        Raises:  N/A
        """
        if self.args.debug and not self.args.quiet:
            header = kwargs.pop("header", "")
            footer = kwargs.pop("footer", "")
            prefix = kwargs.pop("prefix", "")

            if header:
                print(ConsolePrinter._debug_prefix_lines(
                    "{}{}".format(prefix, header)))

            for line in ConsolePrinter._debug_dump(message, prefix=prefix):
                print(line)

            if "data" in kwargs:
                data_header = kwargs.pop("data_header", "")
                data_footer = kwargs.pop("data_footer", "")

                if data_header:
                    print(ConsolePrinter._debug_prefix_lines(
                        "{}{}".format(prefix, data_header)))

                for line in ConsolePrinter._debug_dump(
                    kwargs.pop("data"), prefix=prefix, print_type=True
                ):
                    print(line)

                if data_footer:
                    print(ConsolePrinter._debug_prefix_lines(
                        "{}{}".format(prefix, data_footer)))

            if footer:
                print(ConsolePrinter._debug_prefix_lines(
                    "{}{}".format(prefix, footer)))

    @staticmethod
    def _debug_prefix_lines(line: str) -> str:
        """Helper for debug."""
        return "DEBUG:  {}".format(str(line).replace("\n", "\nDEBUG:  "))

    @staticmethod
    def _debug_get_anchor(data: Any) -> str:
        """Helper for debug."""
        return ("&{}".format(data.anchor.value)
                if (hasattr(data, "anchor")
                    and hasattr(data.anchor, "value")
                    and data.anchor.value is not None)
                else "")

    @staticmethod
    def _debug_get_tag(data: Any) -> str:
        """Helper for debug."""
        return str(data.tag.value
                if (hasattr(data, "tag")
                    and hasattr(data.tag, "value")
                    and data.tag.value is not None)
                else "")

    @staticmethod
    def _debug_dump(data: Any, **kwargs) -> Generator[str, None, None]:
        """Helper for debug."""
        prefix = kwargs.pop("prefix", "")
        if isinstance(data, dict):
            for line in ConsolePrinter._debug_dict(
                data, prefix=prefix, **kwargs
            ):
                yield line
        elif isinstance(data, (list, tuple, deque)):
            for line in ConsolePrinter._debug_list(
                data, prefix=prefix, **kwargs
            ):
                yield line
        elif isinstance(data, (set, CommentedSet)):
            for line in ConsolePrinter._debug_set(
                data, prefix=prefix, **kwargs
            ):
                yield line
        elif isinstance(data, NodeCoords):
            for line in ConsolePrinter._debug_node_coord(
                data, prefix=prefix, **kwargs
            ):
                yield line
        else:
            yield ConsolePrinter._debug_scalar(data, prefix=prefix, **kwargs)

    @staticmethod
    def _debug_scalar(data: Any, **kwargs) -> str:
        """Helper for debug."""
        prefix = kwargs.pop("prefix", "")
        print_anchor = kwargs.pop("print_anchor", True)
        print_tag = kwargs.pop("print_tag", True)
        print_type = kwargs.pop("print_type", False)
        dtype: str = str(type(data)) if print_type else ""
        anchor_prefix = ""
        print_prefix = prefix

        if print_anchor:
            anchor = ConsolePrinter._debug_get_anchor(data)
            if anchor:
                anchor_prefix = "({})".format(anchor)

        if print_tag:
            if isinstance(data, TaggedScalar):
                tag_prefix = "{}{}<{}>".format(
                    print_prefix, anchor_prefix, data.tag.value)
                return ConsolePrinter._debug_scalar(
                        data.value, prefix=tag_prefix,
                        print_anchor=False, print_tag=False, print_type=True)

        # The "true" type of the value is nested in TaggedScalar.value
        if isinstance(data, TaggedScalar):
            dtype = "{}({})".format(dtype, type(data.value))

        # Report fold points, if present
        if hasattr(data, "fold_pos"):
            dtype += ",folded@{}".format(data.fold_pos)

        print_prefix += anchor_prefix

        if isinstance(data, AnchoredDate):
            print_line = data.date().isoformat()
        elif isinstance(data, AnchoredTimeStamp):
            # Import loop occurs when this import is moved to the top because
            # NodeCoords uses Nodes which uses NodeCoords
            #pylint: disable=import-outside-toplevel
            from yamlpath.common.nodes import Nodes
            print_line = Nodes.get_timestamp_with_tzinfo(data).isoformat()
        else:
            print_line = str(data).replace("\n", "\n{}".format(print_prefix))

        return ConsolePrinter._debug_prefix_lines(
            "{}{}{}".format(print_prefix, print_line, dtype))

    @staticmethod
    def _debug_node_coord(
        data: NodeCoords, **kwargs
    ) -> Generator[str, None, None]:
        """Helper method for debug."""
        prefix = kwargs.pop("prefix", "")
        path_prefix = "{}(path)".format(prefix)
        segment_prefix = "{}(segment)".format(prefix)
        node_prefix = "{}(node)".format(prefix)
        parent_prefix = "{}(parent)".format(prefix)
        parentref_prefix = "{}(parentref)".format(prefix)
        ancestry_prefix = "{}(ancestry)".format(prefix)

        for line in ConsolePrinter._debug_dump(data.path, prefix=path_prefix):
            yield line

        for line in ConsolePrinter._debug_dump(
            data.path_segment, prefix=segment_prefix
        ):
            yield line

        for line in ConsolePrinter._debug_dump(data.node, prefix=node_prefix):
            yield line

        for line in ConsolePrinter._debug_dump(
            data.parent, prefix=parent_prefix
        ):
            yield line

        for line in ConsolePrinter._debug_dump(
            data.parentref, prefix=parentref_prefix
        ):
            yield line

        for line in ConsolePrinter._debug_dump(
            data.ancestry, prefix=ancestry_prefix
        ):
            yield line

    @staticmethod
    def _debug_list(
        data: Union[List[Any], Set[Any], Tuple[Any, ...], Deque[Any]], **kwargs
    ) -> Generator[str, None, None]:
        """Helper for debug."""
        prefix = kwargs.pop("prefix", "")
        print_tag = kwargs.pop("print_tag", True)

        if (print_tag
            and isinstance(data, (CommentedBase, CommentedSet))
            and hasattr(data, "tag")
            and data.tag.value
        ):
            prefix += "<{}>".format(data.tag.value)

        for idx, ele in enumerate(data):
            ele_prefix = "{}[{}]".format(prefix, idx)

            # if (isinstance(data, (CommentedBase, CommentedSet))
            #     and hasattr(ele, "tag")
            #     and ele.tag.value
            # ):
            #     ele_prefix += "<{}>".format(ele.tag.value)

            for line in ConsolePrinter._debug_dump(
                ele, prefix=ele_prefix, print_anchor=True, print_tag=True,
                print_type=True
            ):
                yield line

    @staticmethod
    def _debug_get_kv_anchors(key: Any, value: Any) -> str:
        """Helper for debug."""
        key_anchor = ConsolePrinter._debug_get_anchor(key)
        val_anchor = ConsolePrinter._debug_get_anchor(value)
        display_anchor = ""
        if key_anchor and val_anchor:
            display_anchor = "({},{})".format(key_anchor, val_anchor)
        elif key_anchor:
            display_anchor = "({},_)".format(key_anchor)
        elif val_anchor:
            display_anchor = "(_,{})".format(val_anchor)
        return display_anchor

    @staticmethod
    def _debug_get_kv_tags(key: Any, value: Any) -> str:
        """Helper for debug."""
        key_tag = ConsolePrinter._debug_get_tag(key)
        val_tag = ConsolePrinter._debug_get_tag(value)
        display_tag = ""
        if key_tag and val_tag:
            display_tag = "<{},{}>".format(key_tag, val_tag)
        elif key_tag:
            display_tag = "<{},_>".format(key_tag)
        elif val_tag:
            display_tag = "<_,{}>".format(val_tag)
        return display_tag

    @staticmethod
    def _debug_dict(
        data: Union[Dict, CommentedMap], **kwargs
    ) -> Generator[str, None, None]:
        """Helper for debug."""
        prefix = kwargs.pop("prefix", "")

        local_keys = []
        if isinstance(data, CommentedMap):
            for local_key, _ in data.non_merged_items():
                local_keys.append(local_key)
        else:
            for key in data.keys():
                local_keys.append(key)

        for key, val in data.items():
            display_key = (str(key)
                           if key in local_keys
                           else "<<:{}:>>".format(key))
            display_anchor = ConsolePrinter._debug_get_kv_anchors(key, val)
            display_tag = ConsolePrinter._debug_get_kv_tags(key, val)
            kv_prefix = "{}[{}]{}{}".format(
                prefix, display_key, display_anchor, display_tag)

            for line in ConsolePrinter._debug_dump(
                val, prefix=kv_prefix, print_type=True, print_tag=False,
                print_anchor=False
            ):
                yield line

    @staticmethod
    def _debug_set(
        data: Union[Set, CommentedSet], **kwargs
    ) -> Generator[str, None, None]:
        """Helper for debug."""
        prefix = kwargs.pop("prefix", "")

        for key in data:
            display_anchor = ConsolePrinter._debug_get_kv_anchors(key, None)
            display_tag = ConsolePrinter._debug_get_kv_tags(key, None)
            line = "{{{}}}{}{}".format(
                key, display_anchor, display_tag)
            yield ConsolePrinter._debug_prefix_lines(
                "{}{}{}".format(prefix, line, type(key)))



import pytest

from yamlpath.merger.enums.hashmergeopts import (
	HashMergeOpts)


class Test_merger_enums_hashmergeopts():
	"""Tests for the HashMergeOpts enumeration."""

	def test_get_names(self):
		assert HashMergeOpts.get_names() == [
			"DEEP",
			"LEFT",
			"RIGHT",
		]

	def test_get_choices(self):
		assert HashMergeOpts.get_choices() == [
			"deep",
			"left",
			"right",
		]

	@pytest.mark.parametrize("input,output", [
		("DEEP", HashMergeOpts.DEEP),
		("LEFT", HashMergeOpts.LEFT),
		("RIGHT", HashMergeOpts.RIGHT),
	])
	def test_from_str(self, input, output):
		assert output == HashMergeOpts.from_str(input)

	def test_from_str_nameerror(self):
		with pytest.raises(NameError):
			HashMergeOpts.from_str("NO SUCH NAME")



import pytest
import json
import datetime as dt

import ruamel.yaml as ry
from ruamel.yaml import version_info as ryversion
if ryversion < (0, 17, 22):                   # pragma: no cover
    from yamlpath.patches.timestamp import (
        AnchoredTimeStamp,
        AnchoredDate,
    )  # type: ignore
else:                                         # pragma: no cover
    # Temporarily fool MYPY into resolving the future-case imports
    from ruamel.yaml.timestamp import TimeStamp as AnchoredTimeStamp
    AnchoredDate = AnchoredTimeStamp
    #from ruamel.yaml.timestamp import AnchoredTimeStamp
    # From whence shall come AnchoredDate?

from yamlpath.enums import YAMLValueFormats
from yamlpath.common import Parsers

class Test_common_parsers():
    """Tests for the Parsers helper class."""

    ###
    # get_yaml_data (literal=True)
    ###
    def test_get_yaml_data_literally(self, quiet_logger):
        serialized_yaml = """---
hash:
  key: value

list:
  - ichi
  - ni
  - san
"""
        yaml = Parsers.get_yaml_editor()
        (data, loaded) = Parsers.get_yaml_data(
            yaml, quiet_logger, serialized_yaml,
            literal=True)
        assert loaded == True
        assert data["hash"]["key"] == "value"
        assert data["list"][0] == "ichi"
        assert data["list"][1] == "ni"
        assert data["list"][2] == "san"

    ###
    # get_yaml_multidoc_data (literal=True)
    ###
    def test_get_yaml_multidoc_data_literally(self, quiet_logger):
        serialized_yaml = """---
document: 1st
has: data
...
---
document: 2nd
has: different data
"""
        yaml = Parsers.get_yaml_editor()
        doc_id = 0
        for (data, loaded) in Parsers.get_yaml_multidoc_data(
                yaml, quiet_logger, serialized_yaml,
                literal=True):
            assert loaded == True
            if doc_id == 0:
                document = "1st"
                has = "data"
            else:
                document= "2nd"
                has = "different data"
            doc_id = doc_id + 1

            assert data["document"] == document
            assert data["has"] == has

    ###
    # stringify_dates
    ###
    def test_stringify_complex_data_with_dates(self):
        cdata = ry.comments.CommentedMap({
            "dates": ry.comments.CommentedSeq([
                dt.date(2020, 10, 31),
                dt.date(2020, 11, 3)
            ])
        })
        sdata = Parsers.stringify_dates(cdata)
        assert sdata["dates"][0] == "2020-10-31"
        assert sdata["dates"][1] == "2020-11-03"

    ###
    # jsonify_yaml_data
    ###
    def test_jsonify_complex_ruamel_data(self):
        tagged_tag = "!tagged"
        tagged_value = "tagged value"
        tagged_scalar = ry.scalarstring.PlainScalarString(tagged_value)
        tagged_node = ry.comments.TaggedScalar(tagged_scalar, tag=tagged_tag)

        null_tag = "!null"
        null_value = None
        null_node = ry.comments.TaggedScalar(None, tag=null_tag)

        cdata = ry.comments.CommentedMap({
            "tagged": tagged_node,
            "null": null_node,
            "dates": ry.comments.CommentedSeq([
                dt.date(2020, 10, 31),
                dt.date(2020, 11, 3),
                AnchoredDate(2020, 12, 1),
                AnchoredTimeStamp(2021, 1, 13, 1, 2, 3)
            ]),
            "t_bool": ry.scalarbool.ScalarBoolean(1),
            "f_bool": ry.scalarbool.ScalarBoolean(0)
        })
        jdata = Parsers.jsonify_yaml_data(cdata)
        assert jdata["tagged"] == tagged_value
        assert jdata["null"] == null_value
        assert jdata["dates"][0] == "2020-10-31"
        assert jdata["dates"][1] == "2020-11-03"
        assert jdata["dates"][2] == "2020-12-01"
        assert jdata["dates"][3] == "2021-01-13T01:02:03"
        assert jdata["t_bool"] == 1
        assert jdata["f_bool"] == 0

        jstr = json.dumps(jdata)
        assert jstr == """{"tagged": "tagged value", "null": null, "dates": ["2020-10-31", "2020-11-03", "2020-12-01", "2021-01-13T01:02:03"], "t_bool": true, "f_bool": false}"""

    def test_jsonify_complex_python_data(self):
        cdata = {
            "dates": [
                dt.date(2020, 10, 31),
                dt.date(2020, 11, 3)
            ],
            "bytes": b"abc",
            "t_bool": True,
            "f_bool": False
        }
        jdata = Parsers.jsonify_yaml_data(cdata)
        assert jdata["dates"][0] == "2020-10-31"
        assert jdata["dates"][1] == "2020-11-03"
        assert jdata["t_bool"] == True
        assert jdata["f_bool"] == False

        jstr = json.dumps(jdata)
        assert jstr == """{"dates": ["2020-10-31", "2020-11-03"], "bytes": "b'abc'", "t_bool": true, "f_bool": false}"""



import pytest

from types import SimpleNamespace

from ruamel.yaml.comments import CommentedSeq, CommentedMap
from ruamel.yaml.scalarstring import PlainScalarString
from ruamel.yaml.scalarbool import ScalarBoolean
from ruamel.yaml.scalarfloat import ScalarFloat
from ruamel.yaml.scalarint import ScalarInt

from yamlpath.enums import AnchorMatches, PathSearchMethods, YAMLValueFormats
from yamlpath.path import SearchTerms
from yamlpath import YAMLPath
from yamlpath.func import (
    append_list_element,
    build_next_node,
    clone_node,
    create_searchterms_from_pathattributes,
    ensure_escaped,
    escape_path_section,
    get_node_anchor,
    get_yaml_data,
    get_yaml_editor,
    get_yaml_multidoc_data,
    make_float_node,
    make_new_node,
    search_anchor,
    search_matches,
    stringify_dates,
    wrap_type,
)

from tests.conftest import create_temp_yaml_file


@pytest.fixture
def force_ruamel_load_keyboardinterrupt(monkeypatch):
    from ruamel.yaml import YAML as break_class

    def fake_load(*args, **kwargs):
        raise KeyboardInterrupt

    monkeypatch.setattr(break_class, "load", fake_load)
    monkeypatch.setattr(break_class, "load_all", fake_load)

class Test_func():
    ###
    # build_next_node
    ###
    def test_build_next_node(self):
        yaml_path = YAMLPath("/test/path")
        next_node = build_next_node(yaml_path, 0)
        assert isinstance(next_node, dict)

    ###
    # get_yaml_editor
    ###
    def test_get_yaml_editor(self):
        assert get_yaml_editor()

    def test_get_json_editor(self):
        assert get_yaml_editor(explode_aliases=True)


    ###
    # get_yaml_data
    ###
    def test_get_yaml_data_keyboardinterrupt_error(
        self, capsys, quiet_logger, tmp_path_factory,
        force_ruamel_load_keyboardinterrupt
    ):
        yp = get_yaml_editor()
        content = """---
        no: ''
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        (yaml_data, doc_loaded) = get_yaml_data(yp, quiet_logger, yaml_file)
        assert doc_loaded == False
        captured = capsys.readouterr()
        assert -1 < captured.err.find("keyboard interrupt")

    def test_get_yaml_data_filenotfound_error(
        self, capsys, quiet_logger,
        force_ruamel_load_keyboardinterrupt
    ):
        yp = get_yaml_editor()
        (yaml_data, doc_loaded) = get_yaml_data(yp, quiet_logger, "no-such.file")
        assert doc_loaded == False
        captured = capsys.readouterr()
        assert -1 < captured.err.find("File not found")

    def test_get_yaml_data_parser_error(
        self, capsys, quiet_logger,
        imparsible_yaml_file
    ):
        yp = get_yaml_editor()
        (yaml_data, doc_loaded) = get_yaml_data(yp, quiet_logger, imparsible_yaml_file)
        assert doc_loaded == False
        captured = capsys.readouterr()
        assert -1 < captured.err.find("YAML parsing error")

    def test_get_yaml_data_composition_error(
        self, capsys, quiet_logger,
        badcmp_yaml_file
    ):
        yp = get_yaml_editor()
        (yaml_data, doc_loaded) = get_yaml_data(yp, quiet_logger, badcmp_yaml_file)
        assert doc_loaded == False
        captured = capsys.readouterr()
        assert -1 < captured.err.find("YAML composition error")

    def test_get_yaml_data_construction_error(
        self, capsys, quiet_logger, tmp_path_factory
    ):
        yp = get_yaml_editor()
        content = """---
        missing:
          <<:
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        (yaml_data, doc_loaded) = get_yaml_data(yp, quiet_logger, yaml_file)
        assert doc_loaded == False
        captured = capsys.readouterr()
        assert -1 < captured.err.find("YAML construction error")

    def test_get_yaml_data_syntax_error(
        self, capsys, quiet_logger, tmp_path_factory, badsyntax_yaml_file
    ):
        yp = get_yaml_editor()
        (yaml_data, doc_loaded) = get_yaml_data(yp, quiet_logger, badsyntax_yaml_file)
        assert doc_loaded == False
        captured = capsys.readouterr()
        assert -1 < captured.err.find("YAML syntax error")

    def test_get_yaml_data_duplicatekey_error(
        self, capsys, quiet_logger, tmp_path_factory
    ):
        yp = get_yaml_editor()
        content = """---
        key: value1
        key: value2
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        (yaml_data, doc_loaded) = get_yaml_data(yp, quiet_logger, yaml_file)
        assert doc_loaded == False
        captured = capsys.readouterr()
        assert -1 < captured.err.find("Duplicate Hash key detected")

    def test_get_yaml_data_duplicateanchor_error(
        self, capsys, quiet_logger, tmp_path_factory
    ):
        yp = get_yaml_editor()
        content = """---
        aliases:
          - &anchor value1
          - &anchor value2
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        (yaml_data, doc_loaded) = get_yaml_data(yp, quiet_logger, yaml_file)
        assert doc_loaded == False
        captured = capsys.readouterr()
        assert -1 < captured.err.find("Duplicate YAML Anchor detected")


    ###
    # get_yaml_multidoc_data
    ###
    def test_get_yaml_multidoc_data_keyboardinterrupt_error(
        self, capsys, quiet_logger, tmp_path_factory,
        force_ruamel_load_keyboardinterrupt
    ):
        yp = get_yaml_editor()
        content = """---
        no: ''
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        docs_yielded = 0
        for doc in get_yaml_multidoc_data(yp, quiet_logger, yaml_file):
            if not doc and not isinstance(doc, bool):
                docs_yielded += 1
        assert docs_yielded == 0
        captured = capsys.readouterr()
        assert -1 < captured.err.find("keyboard interrupt")

    def test_get_yaml_multidoc_data_filenotfound_error(
        self, capsys, quiet_logger,
        force_ruamel_load_keyboardinterrupt
    ):
        yp = get_yaml_editor()
        docs_yielded = 0
        for doc in get_yaml_multidoc_data(yp, quiet_logger, "no-such.file"):
            if not doc and not isinstance(doc, bool):
                docs_yielded += 1
        assert docs_yielded == 0
        captured = capsys.readouterr()
        assert -1 < captured.err.find("File not found")

    def test_get_yaml_multidoc_data_parser_error(
        self, capsys, quiet_logger,
        imparsible_yaml_file
    ):
        yp = get_yaml_editor()
        docs_yielded = 0
        for doc in get_yaml_multidoc_data(yp, quiet_logger, imparsible_yaml_file):
            if not doc and not isinstance(doc, bool):
                docs_yielded += 1
        assert docs_yielded == 0
        captured = capsys.readouterr()
        assert -1 < captured.err.find("YAML parsing error")

    def test_get_yaml_multidoc_data_composition_error(
        self, capsys, quiet_logger,
        badcmp_yaml_file
    ):
        yp = get_yaml_editor()
        docs_yielded = 0
        for doc in get_yaml_multidoc_data(yp, quiet_logger, badcmp_yaml_file):
            if not doc and not isinstance(doc, bool):
                docs_yielded += 1
        assert docs_yielded == 0
        captured = capsys.readouterr()
        assert -1 < captured.err.find("YAML composition error")

    def test_get_yaml_multidoc_data_construction_error(
        self, capsys, quiet_logger, tmp_path_factory
    ):
        yp = get_yaml_editor()
        content = """---
        missing:
          <<:
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        docs_yielded = 0
        for doc in get_yaml_multidoc_data(yp, quiet_logger, yaml_file):
            if not doc and not isinstance(doc, bool):
                docs_yielded += 1
        assert docs_yielded == 0
        captured = capsys.readouterr()
        assert -1 < captured.err.find("YAML construction error")

    def test_get_yaml_multidoc_data_syntax_error(
        self, capsys, quiet_logger, tmp_path_factory, badsyntax_yaml_file
    ):
        yp = get_yaml_editor()
        docs_yielded = 0
        for doc in get_yaml_multidoc_data(yp, quiet_logger, badsyntax_yaml_file):
            if not doc and not isinstance(doc, bool):
                docs_yielded += 1
        assert docs_yielded == 0
        captured = capsys.readouterr()
        assert -1 < captured.err.find("YAML syntax error")

    def test_get_yaml_multidoc_data_duplicatekey_error(
        self, capsys, quiet_logger, tmp_path_factory
    ):
        yp = get_yaml_editor()
        content = """---
        key: value1
        key: value2
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        docs_yielded = 0
        for doc in get_yaml_multidoc_data(yp, quiet_logger, yaml_file):
            if not doc and not isinstance(doc, bool):
                docs_yielded += 1
        assert docs_yielded == 0
        captured = capsys.readouterr()
        assert -1 < captured.err.find("Duplicate Hash key detected")

    def test_get_yaml_multidoc_data_duplicateanchor_error(
        self, capsys, quiet_logger, tmp_path_factory
    ):
        yp = get_yaml_editor()
        content = """---
        aliases:
          - &anchor value1
          - &anchor value2
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        docs_yielded = 0
        for doc in get_yaml_multidoc_data(yp, quiet_logger, yaml_file):
            if not doc and not isinstance(doc, bool):
                docs_yielded += 1
        assert docs_yielded == 0
        captured = capsys.readouterr()
        assert -1 < captured.err.find("Duplicate YAML Anchor detected")


    ###
    # append_list_element
    ###
    def test_anchorless_list_element_error(self):
        with pytest.raises(ValueError) as ex:
            append_list_element({}, YAMLPath("foo"), "bar")
        assert -1 < str(ex.value).find("Impossible to add an Anchor")


    ###
    # wrap_type
    ###
    @pytest.mark.parametrize("value,checktype", [
        ([], CommentedSeq),
        ({}, CommentedMap),
        ("", PlainScalarString),
        (1, ScalarInt),
        (1.1, ScalarFloat),
        (True, ScalarBoolean),
        (SimpleNamespace(), SimpleNamespace),
    ])
    def test_wrap_type(self, value, checktype):
        assert isinstance(wrap_type(value), checktype)


    ###
    # clone_node
    ###
    def test_clone_node(self):
        test_val = "test"
        assert test_val == clone_node(test_val)

        anchor_val = PlainScalarString(test_val, anchor="anchor")
        assert anchor_val == clone_node(anchor_val)


    ###
    # make_new_node
    ###
    @pytest.mark.parametrize("source,value,check,vformat", [
        ("", " ", " ", YAMLValueFormats.BARE),
        ("", '" "', '" "', YAMLValueFormats.DQUOTE),
        ("", "' '", "' '", YAMLValueFormats.SQUOTE),
        ("", " ", " ", YAMLValueFormats.FOLDED),
        ("", " ", " ", YAMLValueFormats.LITERAL),
        (True, False, False, YAMLValueFormats.BOOLEAN),
        (True, "no", False, YAMLValueFormats.BOOLEAN),
        (1.1, 1.2, 1.2, YAMLValueFormats.FLOAT),
        (ScalarFloat(1.1, anchor="test"), 1.2, 1.2, YAMLValueFormats.FLOAT),
        (1, 2, 2, YAMLValueFormats.INT),
    ])
    def test_make_new_node(self, source, value, check, vformat):
        assert check == make_new_node(source, value, vformat)

    @pytest.mark.parametrize("source,value,vformat,etype,estr", [
        ("", " ", "DNF", NameError, "Unknown YAML Value Format"),
        (1.1, "4F", YAMLValueFormats.FLOAT, ValueError, "cannot be cast to a floating-point number"),
        (1, "4F", YAMLValueFormats.INT, ValueError, "cannot be cast to an integer number"),
    ])
    def test_make_new_node_errors(self, source, value, vformat, etype, estr):
        with pytest.raises(etype) as ex:
            value == make_new_node(source, value, vformat)
        assert -1 < str(ex.value).find(estr)


    ###
    # ensure_escaped
    ###
    def test_ensure_escaped(self):
        assert r"\(\)" == ensure_escaped("()", "(", ")")


    ###
    # escape_path_section
    ###
    def test_escape_path_section(self):
        from yamlpath.enums.pathseperators import PathSeperators
        assert r"a\\b\.c\(\)\[\]\^\$\%\ \'\"" == escape_path_section("a\\b.c()[]^$% '\"", PathSeperators.DOT)


    ###
    # create_searchterms_from_pathattributes
    ###
    def test_create_searchterms_from_pathattributes(self):
        st = SearchTerms(False, PathSearchMethods.EQUALS, ".", "key")
        assert str(st) == str(create_searchterms_from_pathattributes(st))

        with pytest.raises(AttributeError):
            _ = create_searchterms_from_pathattributes("nothing-to-see-here")


    ###
    # search_anchor
    ###
    def test_search_anchor(self):
        anchor_value = "anchor_name"
        node = PlainScalarString("anchored value", anchor=anchor_value)
        terms = SearchTerms(False, PathSearchMethods.CONTAINS, ".", "name")
        seen_anchors = []
        search_anchors = True
        include_aliases = True
        assert search_anchor(node, terms, seen_anchors, search_anchors=search_anchors, include_aliases=include_aliases) == AnchorMatches.MATCH


    ###
    # search_matches
    ###
    def test_search_matches(self):
        method = PathSearchMethods.CONTAINS
        needle = "a"
        haystack = "parents"
        assert search_matches(method, needle, haystack) == True


    ###
    # stringify_dates
    ###
    def test_stringify_dates(self):
        from datetime import date
        yaml_safe_data = CommentedMap({ "string": "Value", "number": 1, "date": date(2020, 10, 19) })
        json_safe_data = { "string": "Value", "number": 1, "date": "2020-10-19" }
        assert stringify_dates(yaml_safe_data) == json_safe_data


    ###
    # make_float_node
    ###
    def test_make_float(self):
        assert isinstance(make_float_node(3.14159265385), ScalarFloat)


    ###
    # get_node_anchor
    ###
    def test_get_node_anchor(self):
        anchor_value = "anchored"
        node = PlainScalarString("value")
        node.yaml_set_anchor(anchor_value)
        assert get_node_anchor(node) == anchor_value



import pytest

from yamlpath.merger.enums.multidocmodes import MultiDocModes


class Test_merger_enums_multidocmodes():
	"""Tests for the MultiDocModes enumeration."""

	def test_get_names(self):
		assert MultiDocModes.get_names() == [
			"CONDENSE_ALL",
			"MERGE_ACROSS",
			"MATRIX_MERGE",
		]

	def test_get_choices(self):
		assert MultiDocModes.get_choices() == [
			"condense_all",
			"matrix_merge",
			"merge_across",
		]

	@pytest.mark.parametrize("input,output", [
		("CONDENSE_ALL", MultiDocModes.CONDENSE_ALL),
		("MERGE_ACROSS", MultiDocModes.MERGE_ACROSS),
		("MATRIX_MERGE", MultiDocModes.MATRIX_MERGE),
	])
	def test_from_str(self, input, output):
		assert output == MultiDocModes.from_str(input)

	def test_from_str_nameerror(self):
		with pytest.raises(NameError):
			MultiDocModes.from_str("NO SUCH NAME")



"""
Implements the ArrayMergeOpts enumeration.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto
from typing import List


class ArrayMergeOpts(Enum):
    """
    Supported Array (AKA: List) Merge Options.

    Options include:

    `ALL`
        All RHS Arrays elements are appended to LHS Arrays (no deduplication).

    `LEFT`
        LHS Arrays are not overwritten/appended by RHS Arrays (no merge).

    `RIGHT`
        RHS Arrays fully replace LHS Arrays (no merge).

    `UNIQUE`
        Only unique RHS Array elements are appended to LHS Arrays (merge).
    """

    ALL = auto()
    LEFT = auto()
    RIGHT = auto()
    UNIQUE = auto()

    @staticmethod
    def get_names() -> List[str]:
        """
        Get all upper-cased entry names for this enumeration.

        Parameters:  N/A

        Returns:  (List[str]) Upper-case names from this enumeration

        Raises:  N/A
        """
        return [entry.name.upper() for entry in ArrayMergeOpts]

    @staticmethod
    def get_choices() -> List[str]:
        """
        Get all entry names with symbolic representations for this enumeration.

        All returned entries are lower-cased.

        Parameters:  N/A

        Returns:  (List[str]) Lower-case names and symbols from this
            enumeration

        Raises:  N/A
        """
        names = [l.lower() for l in ArrayMergeOpts.get_names()]
        choices = list(set(names))
        choices.sort()
        return choices

    @staticmethod
    def from_str(name: str) -> "ArrayMergeOpts":
        """
        Convert a string value to a value of this enumeration, if valid.

        Parameters:
            1. name (str) The name to convert

        Returns:  (ArrayMergeOpts) the converted enumeration value

        Raises:
            - `NameError` when name doesn't match any enumeration values
        """
        check: str = str(name).upper()
        if check in ArrayMergeOpts.get_names():
            return ArrayMergeOpts[check]
        raise NameError(
            "ArrayMergeOpts has no such item:  {}"
            .format(name))



import pytest

from tests.conftest import (
    create_temp_yaml_file,
    requireseyaml,
    old_eyaml_keys,
    new_eyaml_keys,
)

class Test_eyaml_rotate_keys():
    """Tests for the eyaml-rotate-keys command-line interface."""
    command = "eyaml-rotate-keys"

    def test_no_options(self, script_runner):
        result = script_runner.run(self.command)
        assert not result.success, result.stderr
        assert "usage: {}".format(self.command) in result.stderr

    def test_duplicate_keys(self, script_runner):
        bunk_key = "/does/not/exist/on-most/systems"
        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(bunk_key),
            "--newpublickey={}".format(bunk_key),
            "--oldprivatekey={}".format(bunk_key),
            "--oldpublickey={}".format(bunk_key),
            bunk_key
        )
        assert not result.success, result.stderr
        assert "The new and old EYAML keys must be different." in result.stderr

    def test_bad_keys(self, script_runner):
        bunk_file = "/does/not/exist/on-most/systems"
        bunk_old_key = "/does/not/exist/on-most/systems/old"
        bunk_new_key = "/does/not/exist/on-most/systems/new"
        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(bunk_new_key),
            "--newpublickey={}".format(bunk_new_key),
            "--oldprivatekey={}".format(bunk_old_key),
            "--oldpublickey={}".format(bunk_old_key),
            bunk_file
        )
        assert not result.success, result.stderr
        assert "EYAML key is not a readable file:" in result.stderr

    @requireseyaml
    def test_no_yaml_files(self, script_runner, old_eyaml_keys, new_eyaml_keys):
        bunk_file = "/does/not/exist/on-most/systems"
        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(new_eyaml_keys[0]),
            "--newpublickey={}".format(new_eyaml_keys[1]),
            "--oldprivatekey={}".format(old_eyaml_keys[0]),
            "--oldpublickey={}".format(old_eyaml_keys[1]),
            bunk_file
        )
        assert not result.success, result.stderr
        assert "Not a file:" in result.stderr

    @requireseyaml
    def test_good_multi_replacements(self, script_runner, tmp_path_factory, old_eyaml_keys, new_eyaml_keys, quiet_logger):
        from yamlpath.func import unwrap_node_coords
        from yamlpath.common import Parsers
        from yamlpath import Processor
        from yamlpath.eyaml import EYAMLProcessor

        simple_content = """---
encrypted_string: ENC[PKCS7,MIIBiQYJKoZIhvcNAQcDoIIBejCCAXYCAQAxggEhMIIBHQIBADAFMAACAQEwDQYJKoZIhvcNAQEBBQAEggEAHA4rPcTzvgzPLtnGz3yoyX/kVlQ5TnPXcScXK2bwjguGZLkuzv/JVPAsOm4t6GlnROpy4zb/lUMHRJDChJhPLrSj919B8//huoMgw0EU5XTcaN6jeDDjL+vhjswjvLFOux66UwvMo8sRci/e2tlFiam8VgxzV0hpF2qRrL/l84V04gL45kq4PCYDWrJNynOwYVbSIF+qc5HaF25H8kHq1lD3RB6Ob/J942Q7k5Qt7W9mNm9cKZmxwgtUgIZWXW6mcPJ2dXDB/RuPJJSrLsb1VU/DkhdgxaNzvLCA+MViyoFUkCfHFNZbaHKNkoYXBy7dLmoh/E5tKv99FeG/7CzL3DBMBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBCVU5Mjt8+4dLkoqB9YArfkgCDkdIhXR9T1M4YYa1qTE6by61VPU3g1aMExRmo4tNZ8FQ==]
encrypted_block: >
  ENC[PKCS7,MIIBeQYJKoZIhvcNAQcDoIIBajCCAWYCAQAxggEhMIIBHQIBADAFMAACAQEw
  DQYJKoZIhvcNAQEBBQAEggEAnxQVqyIgRTb/+VP4Q+DLJcnlS8YPouXEW8+z
  it9uwUA02CEPxCEU944GcHpgTY3EEtkm+2Z/jgXI119VMML+OOQ1NkwUiAw/
  wq0vwz2D16X31XzhedQN5FZbfZ1C+2tWSQfCjE0bu7IeHfyR+k2ssD11kNZh
  JDEr2bM2dwOdT0y7VGcQ06vI9gw6UXcwYAgS6FoLm7WmFftjcYiNB+0EJSW0
  VcTn2gveaw9iOQcum/Grby+9Ybs28fWd8BoU+ZWDpoIMEceujNa9okIXNPJO
  jcvv1sgauwJ3RX6WFQIy/beS2RT5EOLhWIZCAQCcgJWgovu3maB7dEUZ0NLG
  OYUR7zA8BgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBAbO16EzQ5/cdcvgB0g
  tpKIgBAEgTLT5n9Jtc9venK0CKso]
"""
        anchored_content = """---
aliases:
  - &blockStyle >
    ENC[PKCS7,MIIBiQYJKoZIhvcNAQcDoIIBejCCAXYCAQAxggEhMIIBHQIBADAFMAACAQEw
    DQYJKoZIhvcNAQEBBQAEggEArvk6OYa1gACTdrWq2SpCrtGRlc61la5AGU7L
    aLTyKfqD9vqx71RDjobfOF96No07kLsEpoAJ+LKKHNjdG6kjvpGPmttj9Dkm
    XVoU6A+YCmm4iYFKD/NkoSOEyAkoDOXSqdjrgt0f37GefEsXt6cqAavDpUJm
    pmc0KI4TCG5zpfCxqttMs+stOY3Y+0WokkulQujZ7K3SdWUSHIysgMrWiect
    Wdg5unxN1A/aeyvhgvYSNPjU9KBco7SDnigSs9InW/QghJFrZRrDhTp1oTUc
    qK5lKvaseHkVGi91vPWeLQxZt1loJB5zL6j5BxMbvRfJK+wc3ax2u4x8WTAB
    EurCwzBMBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBAwcy7jvcOGcMfLEtug
    LEXbgCBkocdckuDe14mVGmUmM++xN34OEVRCeGVWWUnWq1DJ4Q==]
  - &stringStyle ENC[PKCS7,MIIBiQYJKoZIhvcNAQcDoIIBejCCAXYCAQAxggEhMIIBHQIBADAFMAACAQEwDQYJKoZIhvcNAQEBBQAEggEAIu44u62q5sVfzC7kytLi2Z/EzH2DKr4vDsoqDBeSZ71aRku/uSrjyiO4lyoq9Kva+eBAyjBay5fnqPVBaU3Rud2pdEoZEoyofi02jn4hxUKpAO1W0AUgsQolGe53qOdM4U8RbwnTR0gr3gp2mCd18pH3SRMP9ryrsBAxGzJ6mR3RgdZnlTlqVGXCeWUeVpbH+lcHw3uvd+o/xkvJ/3ypxz+rWILiAZ3QlCirzn/qb2fHuKf3VBh8RVFuQDaM5voajZlgjD6KzNCsbATOqOA6eJI4j0ngPdDlIjGHAnahuyluQ5f5SIaIjLC+ZeCOfIYni0MQ+BHO0JNbccjq2Unb7TBMBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBCYmAI0Ao3Ok1cSmVw0SgQGgCBK62z1r5RfRjf1xKfqDxTsGUHfsUmM3EjGJfnWzCRvuQ==]
block: *blockStyle
string: *stringStyle
yet_another:
  'more.complex.child': *blockStyle
"""
        simple_file = create_temp_yaml_file(tmp_path_factory, simple_content)
        anchored_file = create_temp_yaml_file(tmp_path_factory, anchored_content)

        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(new_eyaml_keys[0]),
            "--newpublickey={}".format(new_eyaml_keys[1]),
            "--oldprivatekey={}".format(old_eyaml_keys[0]),
            "--oldpublickey={}".format(old_eyaml_keys[1]),
            simple_file,
            anchored_file
        )
        assert result.success, result.stderr

        with open(simple_file, 'r') as fhnd:
            simple_data = fhnd.read()

        with open(anchored_file, 'r') as fhnd:
            anchored_data = fhnd.read()
            # print("=== DEBUG: START anchored_data ===")
            # print(anchored_data)
            # print("=== DEBUG: STOP anchored_data ===")

        assert not simple_data == simple_content
        assert not anchored_data == anchored_content

        # Verify that block and string formatting is correct
        yaml = Parsers.get_yaml_editor()

        (yaml_rotated_data, doc_loaded) = Parsers.get_yaml_data(
            yaml, quiet_logger,
            anchored_data, literal=True)
        if not doc_loaded:
            # An error message has already been logged
            assert False, "Rotated anchored data failed to load (verify formatting)"

        source_processor = Processor(quiet_logger, yaml_rotated_data)
        for node in source_processor.get_nodes('/block', mustexist=True):
            assert not '  ' in unwrap_node_coords(node)


        # Test that the pre- and post-rotated values are identical
        (yaml_anchored_data, doc_loaded) = Parsers.get_yaml_data(
            yaml, quiet_logger,
            anchored_content, literal=True)
        if not doc_loaded:
            # An error message has already been logged
            assert False, "Original anchored data failed to load (verify values)"

        (yaml_rotated_data, doc_loaded) = Parsers.get_yaml_data(
            yaml, quiet_logger,
            anchored_data, literal=True)
        if not doc_loaded:
            # An error message has already been logged
            assert False, "Rotated anchored data failed to load (verify values)"

        source_processor = EYAMLProcessor(
            quiet_logger, yaml_anchored_data,
            privatekey=old_eyaml_keys[0],
            publickey=old_eyaml_keys[1])
        for node in source_processor.get_eyaml_values(
            '/block', True
        ):
            assert unwrap_node_coords(node) == 'This is a test value.'

        rotated_processor = EYAMLProcessor(
            quiet_logger, yaml_rotated_data,
            privatekey=new_eyaml_keys[0],
            publickey=new_eyaml_keys[1])
        for node in rotated_processor.get_eyaml_values(
            '/block', True
        ):
            assert unwrap_node_coords(node) == 'This is a test value.'

    def test_yaml_parsing_error(self, script_runner, imparsible_yaml_file, old_eyaml_keys, new_eyaml_keys):
        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(new_eyaml_keys[0]),
            "--newpublickey={}".format(new_eyaml_keys[1]),
            "--oldprivatekey={}".format(old_eyaml_keys[0]),
            "--oldpublickey={}".format(old_eyaml_keys[1]),
            imparsible_yaml_file
        )
        assert not result.success, result.stderr
        assert "YAML parsing error" in result.stderr

    def test_yaml_syntax_error(self, script_runner, badsyntax_yaml_file, old_eyaml_keys, new_eyaml_keys):
        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(new_eyaml_keys[0]),
            "--newpublickey={}".format(new_eyaml_keys[1]),
            "--oldprivatekey={}".format(old_eyaml_keys[0]),
            "--oldpublickey={}".format(old_eyaml_keys[1]),
            badsyntax_yaml_file
        )
        assert not result.success, result.stderr
        assert "YAML syntax error" in result.stderr

    def test_yaml_composition_error(self, script_runner, badcmp_yaml_file, old_eyaml_keys, new_eyaml_keys):
        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(new_eyaml_keys[0]),
            "--newpublickey={}".format(new_eyaml_keys[1]),
            "--oldprivatekey={}".format(old_eyaml_keys[0]),
            "--oldpublickey={}".format(old_eyaml_keys[1]),
            badcmp_yaml_file
        )
        assert not result.success, result.stderr
        assert "YAML composition error" in result.stderr

    def test_corrupted_eyaml_value(self, script_runner, tmp_path_factory, old_eyaml_keys, new_eyaml_keys):
        content = """---
        key: >
            ENC[PKCS7,MII ... corrupted-value ...
            DBAEqBBAwcy7jvcOGcMfLEtugGVWWUnWq1DJ4Q==]
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(new_eyaml_keys[0]),
            "--newpublickey={}".format(new_eyaml_keys[1]),
            "--oldprivatekey={}".format(old_eyaml_keys[0]),
            "--oldpublickey={}".format(old_eyaml_keys[1]),
            yaml_file
        )
        assert not result.success, result.stderr
        assert "Unable to decrypt value!" in result.stderr

    def test_bad_recryption_key(self, script_runner, tmp_path_factory, old_eyaml_keys, new_eyaml_keys):
        content = """---
        key: >
          ENC[PKCS7,MIIBmQYJKoZIhvcNAQcDoIIBijCCAYYCAQAxggEhMIIBHQIBADAFMAACAQEw
          DQYJKoZIhvcNAQEBBQAEggEAs+8byhxSVFkzPAfGFazxsifEJQO3RH5MNf2g
          o/x0oh+y1SB6bwB/lPtCBnCwDgKUKR8VzqWM8sTYkLTkSWq5BxS+Hix0zL1u
          zqdzNbuFDNS3PoUM4XaBRPOhGL/xUGc8EuUmdc3RaGRqisZvqACAMDDMme5m
          sCJVHw/QC//hAH6zrPmPA8D5S6ibMHGURifqTmLvi1BxxzMIWXWBmRpadAaq
          nYqhYsI/IWyQBmF7OAwsREREu+qEiDDBOS5IchDcDnlxtoooB5xin4HDS9ED
          MJMlKfpB1FCNtrC4RJz4uqFuwvX482cct3TtS+/UrPLP7rm6EILs7QSQGsdM
          G+8k8DBcBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBD+Bx88oA/j57i5UB5U
          BHEogDDpFaKbtiGSTxOK44MpjLOGCZ4ME6lJz5EYVJQ3VJw95z98mvj6CgzL
          NI/TSIF7M9U=]
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(new_eyaml_keys[1]),
            "--newpublickey={}".format(new_eyaml_keys[0]),
            "--oldprivatekey={}".format(old_eyaml_keys[0]),
            "--oldpublickey={}".format(old_eyaml_keys[1]),
            yaml_file
        )
        assert not result.success, result.stderr
        assert "unable to encrypt" in result.stderr or "cannot be run due to exit code:  1" in result.stderr

    def test_backup_file(self, script_runner, tmp_path_factory, old_eyaml_keys, new_eyaml_keys):
        import os

        content = """---
        key: >
          ENC[PKCS7,MIIBiQYJKoZIhvcNAQcDoIIBejCCAXYCAQAxggEhMIIBHQIBADAFMAACAQEw
          DQYJKoZIhvcNAQEBBQAEggEAPGA1g1Wx50RK8F/Y118w1VT/SnCa7PMfN2OM
          d82vGeWXm6INmoURMDWEvBUEFCmGZoOMLVlK3LALtUcPEW1N9ztJTypBrqqI
          1K8L9aZWRNFt7uwsaoHWvk1XjMujP+nn2ZO3OiFYkiWFh0PcFw7cT1TmexB4
          cNbBtNi7oJ88L17/8rbtJW465cWyj0pPCmwo3OvK39JcuJ2xosujNk4u5AUf
          TjWwklk3yjPvjG6AvoS4TK+vkmqUcCkyy0tLZR8Xu+3IzYCq+DYH4QBrrrZf
          pKer9VawzMzxgVXeCgKGEsa3XeSzWtgbyoZVtoBdl3uv2f8rGi5qAlwZ9syO
          Aold9zBMBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBDGUmDGJfp2Iqn7bATf
          r0H9gCBNamGg9iiM92wGcVSkNmGJtVk8yEe3EOVn/QNzQ6v0fw==]
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        backup_file = yaml_file + ".bak"
        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(new_eyaml_keys[0]),
            "--newpublickey={}".format(new_eyaml_keys[1]),
            "--oldprivatekey={}".format(old_eyaml_keys[0]),
            "--oldpublickey={}".format(old_eyaml_keys[1]),
            "--backup",
            yaml_file
        )
        assert result.success, result.stderr
        assert os.path.isfile(backup_file)

        with open(backup_file, 'r') as fhnd:
            filedat = fhnd.read()
        assert filedat == content

    def test_replace_backup_file(self, script_runner, tmp_path_factory, old_eyaml_keys, new_eyaml_keys):
        import os

        content = """---
        key: >
          ENC[PKCS7,MIIBiQYJKoZIhvcNAQcDoIIBejCCAXYCAQAxggEhMIIBHQIBADAFMAACAQEw
          DQYJKoZIhvcNAQEBBQAEggEAPGA1g1Wx50RK8F/Y118w1VT/SnCa7PMfN2OM
          d82vGeWXm6INmoURMDWEvBUEFCmGZoOMLVlK3LALtUcPEW1N9ztJTypBrqqI
          1K8L9aZWRNFt7uwsaoHWvk1XjMujP+nn2ZO3OiFYkiWFh0PcFw7cT1TmexB4
          cNbBtNi7oJ88L17/8rbtJW465cWyj0pPCmwo3OvK39JcuJ2xosujNk4u5AUf
          TjWwklk3yjPvjG6AvoS4TK+vkmqUcCkyy0tLZR8Xu+3IzYCq+DYH4QBrrrZf
          pKer9VawzMzxgVXeCgKGEsa3XeSzWtgbyoZVtoBdl3uv2f8rGi5qAlwZ9syO
          Aold9zBMBgkqhkiG9w0BBwEwHQYJYIZIAWUDBAEqBBDGUmDGJfp2Iqn7bATf
          r0H9gCBNamGg9iiM92wGcVSkNmGJtVk8yEe3EOVn/QNzQ6v0fw==]
        """
        yaml_file = create_temp_yaml_file(tmp_path_factory, content)
        backup_file = yaml_file + ".bak"
        with open(backup_file, 'w') as fhnd:
            fhnd.write(content + "\nkey2: plain scalar string value")

        result = script_runner.run(
            self.command,
            "--newprivatekey={}".format(new_eyaml_keys[0]),
            "--newpublickey={}".format(new_eyaml_keys[1]),
            "--oldprivatekey={}".format(old_eyaml_keys[0]),
            "--oldpublickey={}".format(old_eyaml_keys[1]),
            "--backup",
            yaml_file
        )
        assert result.success, result.stderr
        assert os.path.isfile(backup_file)

        with open(backup_file, 'r') as fhnd:
            filedat = fhnd.read()
        assert filedat == content



"""
Implements the AnchorMatches enumeration.

Copyright 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto


class AnchorMatches(Enum):
    """
    When performing a search, YAML Anchors and Aliases can be evaluated.

    When they are, these are the possible match results:

    `ALIAS_EXCLUDED`
        The Anchor is a duplicate that has already matched.

    `ALIAS_INCLUDED`
        The Anchor is a duplicate alias and the search parameters permit
        duplicates.

    `MATCH`
        This original Anchor is a match.

    `NO_ANCHOR`
        The given node has no Anchor.

    `NO_MATCH`
        This original Anchor is not a match.

    `UNSEARCHABLE_ALIAS`
        The node references an Anchor via an Alias that has already been seen,
        but the search parameters prohibit searching Anchor names.

    `UNSEARCHABLE_ANCHOR`
        The node has an Anchor that is so-far unique, but the search parameters
        prohibit searching Anchor names.
    """

    ALIAS_EXCLUDED = auto()
    ALIAS_INCLUDED = auto()
    MATCH = auto()
    NO_ANCHOR = auto()
    NO_MATCH = auto()
    UNSEARCHABLE_ALIAS = auto()
    UNSEARCHABLE_ANCHOR = auto()



"""
Defines a custom type for YAML Path segment attributes.

Copyright 2020 William W. Kimball, Jr. MBA MSIS
"""
from typing import Union

from yamlpath.path import CollectorTerms
from yamlpath.path import searchterms


PathAttributes = Union[str, int, CollectorTerms, searchterms.SearchTerms, None]



"""
Implement YAML document Differ.

Copyright 2020, 2021 William W. Kimball, Jr. MBA MSIS
"""
from itertools import zip_longest
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

from ruamel.yaml.comments import CommentedMap, CommentedSeq, CommentedSet

from yamlpath import YAMLPath
from yamlpath.wrappers import ConsolePrinter, NodeCoords
from yamlpath.eyaml import EYAMLProcessor
from .enums import ArrayDiffOpts, AoHDiffOpts, DiffActions
from .diffentry import DiffEntry
from .differconfig import DifferConfig


class Differ:
    """Calculates the difference between two YAML documents."""

    def __init__(
        self, config: DifferConfig, logger: ConsolePrinter, document: Any,
        **kwargs
    ) -> None:
        """
        Instantiate this class into an object.

        Parameters:
        1. logger (ConsolePrinter) Instance of ConsoleWriter or subclass
        2. document (Any) The basis document

        Keyword Arguments:
        * ignore_eyaml_values (bool) Do not decrypt encrypted YAML value for
          comparison

        Returns:  N/A

        Raises:  N/A
        """
        ignore_eyaml = kwargs.pop("ignore_eyaml_values", True)

        self.config: DifferConfig = config
        self.logger: ConsolePrinter = logger
        self._data: Any = document
        self._diffs: List[DiffEntry] = []
        self._ignore_eyaml: bool = ignore_eyaml
        self._eyamlproc = (None
                           if ignore_eyaml
                           else EYAMLProcessor(logger, document, **kwargs))

    def compare_to(self, document: Any) -> None:
        """
        Perform the diff calculation.

        Parameers:
        1. document (Any) The document to compare against

        Returns:  N/A
        """
        self._diffs.clear()
        self.config.prepare(document)
        self._diff_between(YAMLPath(), self._data, document)

    def get_report(self) -> Generator[DiffEntry, None, None]:
        """
        Get the diff report.

        Parameters:  N/A

        Returns:  (Generator[DiffEntry, None, None]) Sorted DiffEntry records
        """
        for entry in sorted(
            self._diffs, key=lambda e: [int(i) for i in e.index.split('.')]
        ):
            yield entry

    def _purge_document(self, path: YAMLPath, data: Any) -> None:
        """
        Record changes necessary to delete every node in the document.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. data (Any) The DOM element under evaluation

        Returns:  N/A
        """
        if isinstance(data, CommentedMap):
            lhs_iteration = -1
            for key, val in data.items():
                lhs_iteration += 1
                next_path = (path +
                    YAMLPath.escape_path_section(key, path.seperator))
                self._diffs.append(
                    DiffEntry(
                        DiffActions.DELETE, next_path, val, None,
                        lhs_parent=data, lhs_iteration=lhs_iteration))
        elif isinstance(data, CommentedSeq):
            for idx, ele in enumerate(data):
                next_path = path + "[{}]".format(idx)
                self._diffs.append(
                    DiffEntry(
                        DiffActions.DELETE, next_path, ele, None,
                        lhs_parent=data, lhs_iteration=idx))
        elif isinstance(data, CommentedSet):
            for idx, ele in enumerate(data):
                next_path = (path +
                    YAMLPath.escape_path_section(ele, path.seperator))
                self._diffs.append(
                    DiffEntry(
                        DiffActions.DELETE, next_path, ele, None,
                        lhs_parent=data, lhs_iteration=idx))
        else:
            if data is not None:
                self._diffs.append(
                    DiffEntry(DiffActions.DELETE, path, data, None)
                )

    def _add_everything(self, path: YAMLPath, data: Any) -> None:
        """
        Record changes necessary to add every node in the document.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. data (Any) The DOM element under evaluation

        Returns:  N/A
        """
        if isinstance(data, CommentedMap):
            rhs_iteration = -1
            for key, val in data.items():
                rhs_iteration += 1
                next_path = (path +
                    YAMLPath.escape_path_section(key, path.seperator))
                self._diffs.append(
                    DiffEntry(
                        DiffActions.ADD, next_path, None, val,
                        rhs_parent=data, rhs_iteration=rhs_iteration))
        elif isinstance(data, CommentedSeq):
            for idx, ele in enumerate(data):
                next_path = path + "[{}]".format(idx)
                self._diffs.append(
                    DiffEntry(
                        DiffActions.ADD, next_path, None, ele,
                        rhs_parent=data, rhs_iteration=idx))
        elif isinstance(data, CommentedSet):
            for idx, ele in enumerate(data):
                next_path = (path +
                    YAMLPath.escape_path_section(ele, path.seperator))
                self._diffs.append(
                    DiffEntry(
                        DiffActions.ADD, next_path, None, ele,
                        rhs_parent=data, rhs_iteration=idx))
        else:
            if data is not None:
                self._diffs.append(
                    DiffEntry(DiffActions.ADD, path, None, data)
                )

    def _diff_scalars(
        self, path: YAMLPath, lhs: Any, rhs: Any, **kwargs
    ) -> None:
        """
        Diff two Scalar values.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document

        Keyword Arguments:  See `DiffEntry`

        Returns:  N/A
        """
        self.logger.debug(
            "Comparing LHS:",
            prefix="Differ::_diff_scalars:  ",
            data=lhs)
        self.logger.debug(
            "Against RHS:",
            prefix="Differ::_diff_scalars:  ",
            data=rhs)

        lhs_val = lhs
        rhs_val = rhs
        if (not self._ignore_eyaml
            and isinstance(self._eyamlproc, EYAMLProcessor)
        ):
            if self._eyamlproc.is_eyaml_value(lhs):
                lhs_val = self._eyamlproc.decrypt_eyaml(lhs)
                lhs = lhs.replace("\r", "").replace(" ", "")
            if self._eyamlproc.is_eyaml_value(rhs):
                rhs_val = self._eyamlproc.decrypt_eyaml(rhs)
                rhs = rhs.replace("\r", "").replace(" ", "")

        if lhs_val == rhs_val:
            self._diffs.append(
                DiffEntry(DiffActions.SAME, path, lhs, rhs, **kwargs)
            )
        else:
            self._diffs.append(
                DiffEntry(DiffActions.CHANGE, path, lhs, rhs, **kwargs)
            )

    def _diff_dicts(
        self, path: YAMLPath, lhs: CommentedMap, rhs: CommentedMap
    ) -> None:
        """
        Diff two dicts.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document
        """
        self.logger.debug(
            "Comparing LHS:",
            prefix="Differ::_diff_dicts:  ",
            data=lhs)
        self.logger.debug(
            "Against RHS:",
            prefix="Differ::_diff_dicts:  ",
            data=rhs)

        # Check first for a difference in YAML Tag
        lhs_tag = lhs.tag.value if hasattr(lhs, "tag") else None
        rhs_tag = rhs.tag.value if hasattr(rhs, "tag") else None
        if lhs_tag != rhs_tag:
            self.logger.debug(
                "Dictionaries have different YAML Tags; {} != {}:".format(
                    lhs_tag, rhs_tag),
                prefix="Differ::_diff_dicts:  ")
            self._diffs.append(
                DiffEntry(
                    DiffActions.DELETE, path, lhs, None, key_tag=lhs_tag))
            self._diffs.append(
                DiffEntry(
                    DiffActions.ADD, path, None, rhs, key_tag=rhs_tag))
            return

        lhs_keys = set(lhs)
        rhs_keys = set(rhs)
        lhs_key_indicies = Differ._get_key_indicies(lhs)
        rhs_key_indicies = Differ._get_key_indicies(rhs)

        self.logger.debug(
            "Got LHS key indicies:",
            prefix="Differ::_diff_dicts:  ",
            data=lhs_key_indicies)
        self.logger.debug(
            "Got RHS key indicies:",
            prefix="Differ::_diff_dicts:  ",
            data=rhs_key_indicies)

        # Look for changes
        for key, val in [
            (key, val) for key, val in rhs.items()
            if key in lhs and key in rhs
        ]:
            next_path = (path +
                YAMLPath.escape_path_section(key, path.seperator))
            self._diff_between(
                next_path, lhs[key], val,
                lhs_parent=lhs, lhs_iteration=lhs_key_indicies[key],
                rhs_parent=rhs, rhs_iteration=rhs_key_indicies[key],
                parentref=key)

        # Look for deleted keys
        for key in lhs_keys - rhs_keys:
            next_path = (path +
                YAMLPath.escape_path_section(key, path.seperator))
            self._diffs.append(
                DiffEntry(
                    DiffActions.DELETE, next_path, lhs[key], None,
                    lhs_parent=lhs, lhs_iteration=lhs_key_indicies[key],
                    rhs_parent=rhs,
                    key_tag=key.tag.value if hasattr(key, "tag") else None))

        # Look for new keys
        for key in rhs_keys - lhs_keys:
            next_path = (path +
                YAMLPath.escape_path_section(key, path.seperator))
            self._diffs.append(
                DiffEntry(
                    DiffActions.ADD, next_path, None, rhs[key],
                    lhs_parent=lhs,
                    rhs_parent=rhs, rhs_iteration=rhs_key_indicies[key],
                    key_tag=key.tag.value if hasattr(key, "tag") else None))

    def _diff_synced_lists(
        self, path: YAMLPath, lhs: CommentedSeq, rhs: CommentedSeq
    ) -> None:
        """
        Diff two synchronized lists.

        A "synchronized" list -- in this context -- is one in which all
        elements that are identical to those of its exemplar list are
        (re)positioned to identical index.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document

        Returns:  N/A
        """
        self.logger.debug("Differ::_diff_synced_lists:  Starting...")
        self.logger.debug(
            "Synchronizing LHS Array elements at YAML Path, {}:"
            .format(path if path else "/"),
            prefix="Differ::_diff_syncd_lists:  ",
            data=lhs)
        self.logger.debug(
            "Synchronizing RHS Array elements at YAML Path, {}:"
            .format(path if path else "/"),
            prefix="Differ::_diff_syncd_lists:  ",
            data=rhs)

        syn_pairs = Differ.synchronize_lists_by_value(lhs, rhs)
        self.logger.debug(
            "Got synchronized pairs of Array elements at YAML Path, {}:"
            .format(path if path else "/"),
            prefix="Differ::_diff_syncd_lists:  ",
            data=syn_pairs)

        for (lidx, lele, ridx, rele) in syn_pairs:
            if lele is None:
                next_path = path + "[{}]".format(ridx)
                diff_action = DiffActions.ADD
                opposite_val = None
                pop_index = -1
                for idx, ele in reversed(list(enumerate(self._diffs))):
                    if (ele.action is DiffActions.DELETE
                        and ele.path == next_path
                    ):
                        pop_index = idx
                        break

                # This YAML Path has ALREADY been recorded as a DELETE.  Since
                # a DELETE->ADD action is really just a CHANGE, remove the
                # conflicting entry and convert this pending ADD to a CHANGE.
                if pop_index > -1:
                    opposite_val = self._diffs.pop(pop_index).lhs
                    diff_action = DiffActions.CHANGE

                self._diffs.append(DiffEntry(
                    diff_action, next_path, opposite_val, rele,
                    lhs_parent=lhs, lhs_iteration=lidx,
                    rhs_parent=rhs, rhs_iteration=ridx))
            elif rele is None:
                next_path = path + "[{}]".format(lidx)
                self._diffs.append(
                    DiffEntry(
                        DiffActions.DELETE, next_path, lele, None,
                        lhs_parent=lhs, lhs_iteration=lidx,
                        rhs_parent=rhs, rhs_iteration=ridx))
            else:
                next_path = path + "[{}]".format(lidx)
                self._diff_between(
                    next_path, lele, rele,
                    lhs_parent=lhs, lhs_iteration=lidx,
                    rhs_parent=rhs, rhs_iteration=ridx,
                    parentref=ridx)

    def _diff_arrays_of_scalars(
        self, path: YAMLPath, lhs: CommentedSeq, rhs: CommentedSeq,
        node_coord: NodeCoords, **kwargs
    ) -> None:
        """
        Diff two lists of scalars.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document
        4. node_coord (NodeCoords) The node being evaluated

        Keyword Parameers:
        * diff_deeply (bool) True = Deeply traverse complex elements; False =
          compare complex elements as-is
        """
        self.logger.debug(
            "Comparing LHS:",
            prefix="Differ::_diff_arrays_of_scalars:  ",
            data=lhs)
        self.logger.debug(
            "Against RHS:",
            prefix="Differ::_diff_arrays_of_scalars:  ",
            data=rhs)

        diff_mode = self.config.array_diff_mode(node_coord)
        if diff_mode is ArrayDiffOpts.VALUE:
            self._diff_synced_lists(path, lhs, rhs)
            return

        idx = 0
        diff_deeply = kwargs.pop("diff_deeply", True)
        for (lele, rele) in zip_longest(lhs, rhs):
            next_path = path + "[{}]".format(idx)
            idx += 1
            if lele is None:
                self._diffs.append(
                    DiffEntry(
                        DiffActions.ADD, next_path, None, rele,
                        lhs_parent=lhs, lhs_iteration=idx,
                        rhs_parent=rhs, rhs_iteration=idx))
            elif rele is None:
                self._diffs.append(
                    DiffEntry(
                        DiffActions.DELETE, next_path, lele, None,
                        lhs_parent=lhs, lhs_iteration=idx,
                        rhs_parent=rhs, rhs_iteration=idx))
            elif diff_deeply:
                self._diff_between(
                    next_path, lele, rele,
                    lhs_parent=lhs, lhs_iteration=idx,
                    rhs_parent=rhs, rhs_iteration=idx,
                    parentref=idx)
            elif lele != rele:
                self._diffs.append(
                    DiffEntry(
                        DiffActions.CHANGE, next_path, lele, rele,
                        lhs_parent=lhs, lhs_iteration=idx,
                        rhs_parent=rhs, rhs_iteration=idx,
                        parentref=idx))

    def _diff_arrays_of_hashes(
        self, path: YAMLPath, lhs: CommentedSeq, rhs: CommentedSeq,
        node_coord: NodeCoords
    ) -> None:
        """
        Diff two lists-of-dictionaries.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document
        4. node_coord (NodeCoords) The node being evaluated

        Returns:  N/A
        """
        self.logger.debug(
            "Comparing LHS:",
            prefix="Differ::_diff_arrays_of_hashes:  ",
            data=lhs)
        self.logger.debug(
            "Against RHS:",
            prefix="Differ::_diff_arrays_of_hashes:  ",
            data=rhs)

        diff_mode = self.config.aoh_diff_mode(node_coord)
        if diff_mode is AoHDiffOpts.POSITION:
            self._diff_arrays_of_scalars(
                path, lhs, rhs, node_coord, diff_deeply=False)
            return
        if diff_mode is AoHDiffOpts.DPOS:
            self._diff_arrays_of_scalars(
                path, lhs, rhs, node_coord, diff_deeply=True)
            return
        if diff_mode is AoHDiffOpts.VALUE:
            self._diff_synced_lists(path, lhs, rhs)
            return
        deep_diff = diff_mode is AoHDiffOpts.DEEP

        self.logger.debug(
            "Synchronizing LHS Array elements at YAML Path, {}:"
            .format(path if path else "/"),
            prefix="Differ::_diff_arrays_of_hashes:  ",
            data=lhs)
        self.logger.debug(
            "Synchronizing RHS Array elements at YAML Path, {}:"
            .format(path if path else "/"),
            prefix="Differ::_diff_arrays_of_hashes:  ",
            data=rhs)

        # Perform either a KEY or DEEP comparison; either way, the elements
        # must first be synchronized based on their identity key values.
        syn_pairs = self.synchronize_lods_by_key(path, lhs, rhs)
        self.logger.debug(
            "Got synchronized pairs of Array elements at YAML Path, {}:"
            .format(path if path else "/"),
            prefix="Differ::_diff_arrays_of_hashes:  ",
            data=syn_pairs)

        for (lidx, lele, ridx, rele) in syn_pairs:
            if lele is None:
                next_path = path + "[{}]".format(ridx)
                self._diffs.append(
                    DiffEntry(
                        DiffActions.ADD, next_path, None, rele,
                        lhs_parent=lhs, lhs_iteration=lidx,
                        rhs_parent=rhs, rhs_iteration=ridx))
            elif rele is None:
                next_path = path + "[{}]".format(lidx)
                self._diffs.append(
                    DiffEntry(
                        DiffActions.DELETE, next_path, lele, None,
                        lhs_parent=lhs, lhs_iteration=lidx,
                        rhs_parent=rhs, rhs_iteration=ridx))
            else:
                if deep_diff:
                    next_path = path + "[{}]".format(ridx)
                    self._diff_between(
                        next_path, lele, rele,
                        lhs_parent=lhs, lhs_iteration=lidx,
                        rhs_parent=rhs, rhs_iteration=ridx,
                        parentref=ridx)
                else:
                    # KEY-based comparisons
                    next_path = path + "[{}]".format(lidx)
                    diff_action = (DiffActions.SAME
                                  if lele == rele
                                  else DiffActions.CHANGE)
                    self._diffs.append(
                        DiffEntry(diff_action, next_path, lele, rele,
                        lhs_parent=lhs, lhs_iteration=lidx,
                        rhs_parent=rhs, rhs_iteration=ridx,
                        parentref=lidx))

    def _diff_lists(
        self, path: YAMLPath, lhs: CommentedSeq, rhs: CommentedSeq, **kwargs
    ) -> None:
        """
        Diff two lists.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document

        Keyword Arguments:
        * rhs_parent (Any) Parent data node of rhs
        * parentref (Any) Reference indicating rhs within rhs_parent

        Returns:  N/A
        """
        self.logger.debug(
            "Comparing LHS:",
            prefix="Differ::_diff_lists:  ",
            data=lhs)
        self.logger.debug(
            "Against RHS:",
            prefix="Differ::_diff_lists:  ",
            data=rhs)

        parent: Any = kwargs.pop("rhs_parent", None)
        parentref: Any = kwargs.pop("parentref", None)
        node_coord = NodeCoords(rhs, parent, parentref)
        if len(rhs) > 0:
            if isinstance(rhs[0], CommentedMap):
                # This list is an Array-of-Hashes
                self._diff_arrays_of_hashes(path, lhs, rhs, node_coord)
            else:
                # This list is an Array-of-Arrays or a simple list of Scalars
                self._diff_arrays_of_scalars(path, lhs, rhs, node_coord)

    # pylint: disable=too-many-locals
    def _diff_sets(
        self, path: YAMLPath, lhs: CommentedSet, rhs: CommentedSet
    ) -> None:
        """
        Diff two sets.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document

        Keyword Arguments:
        * rhs_parent (Any) Parent data node of rhs
        * parentref (Any) Reference indicating rhs within rhs_parent

        Returns:  N/A
        """
        self.logger.debug(
            "Comparing LHS:",
            prefix="Differ::_diff_sets:  ",
            data=lhs)
        self.logger.debug(
            "Against RHS:",
            prefix="Differ::_diff_sets:  ",
            data=rhs)

        # Sets must allways have a tag of !!set lest PyYAML/ruamel.yaml fail to
        # load them as set data.  Keep this code commented until a case is
        # found which proves otherwise.
        # # Check first for a difference in YAML Tag
        # lhs_tag = lhs.tag.value if hasattr(lhs, "tag") else None
        # rhs_tag = rhs.tag.value if hasattr(rhs, "tag") else None
        # if lhs_tag != rhs_tag:
        #     self.logger.debug(
        #         "Sets have different YAML Tags; {} != {}:".format(
        #             lhs_tag, rhs_tag),
        #         prefix="Differ::_diff_sets:  ")
        #     self._diffs.append(
        #         DiffEntry(
        #             DiffActions.DELETE, path, lhs, None, key_tag=lhs_tag))
        #     self._diffs.append(
        #         DiffEntry(
        #             DiffActions.ADD, path, None, rhs, key_tag=rhs_tag))
        #     return

        lhs_keys = set(lhs)
        rhs_keys = set(rhs)
        lhs_key_indicies = Differ._get_key_indicies(lhs)
        rhs_key_indicies = Differ._get_key_indicies(rhs)

        self.logger.debug(
            "Got LHS key indicies:",
            prefix="Differ::_diff_sets:  ",
            data=lhs_key_indicies)
        self.logger.debug(
            "Got RHS key indicies:",
            prefix="Differ::_diff_sets:  ",
            data=rhs_key_indicies)

        # Look for changes
        for key in [
            (key) for key in rhs
            if key in lhs and key in rhs
        ]:
            next_path = (path +
                YAMLPath.escape_path_section(key, path.seperator))
            lhs_ele = None
            rhs_ele = None
            for ele in lhs:
                if ele == key:
                    lhs_ele = ele
                    break
            for ele in rhs:
                if ele == key:
                    rhs_ele = ele
                    break
            self._diff_between(
                next_path, lhs_ele, rhs_ele,
                lhs_parent=lhs, lhs_iteration=lhs_key_indicies[key],
                rhs_parent=rhs, rhs_iteration=rhs_key_indicies[key],
                parentref=key)

        # Look for deleted keys
        for key in lhs_keys - rhs_keys:
            next_path = (path +
                YAMLPath.escape_path_section(key, path.seperator))
            self._diffs.append(
                DiffEntry(
                    DiffActions.DELETE, next_path, key, None,
                    lhs_parent=lhs, lhs_iteration=lhs_key_indicies[key],
                    rhs_parent=rhs,
                    key_tag=key.tag.value if hasattr(key, "tag") else None))

        # Look for new keys
        for key in rhs_keys - lhs_keys:
            next_path = (path +
                YAMLPath.escape_path_section(key, path.seperator))
            self._diffs.append(
                DiffEntry(
                    DiffActions.ADD, next_path, None, key,
                    lhs_parent=lhs,
                    rhs_parent=rhs, rhs_iteration=rhs_key_indicies[key],
                    key_tag=key.tag.value if hasattr(key, "tag") else None))

    def _diff_between(
        self, path: YAMLPath, lhs: Any, rhs: Any, **kwargs
    ) -> None:
        """
        Calculate the differences between two document nodes.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document

        Keyword Arguments:  See _diff_lists() and _diff_scalars()
        """
        self.logger.debug(
            "Comparing LHS:",
            prefix="Differ::_diff_between:  ",
            data=lhs)
        self.logger.debug(
            "Against RHS:",
            prefix="Differ::_diff_between:  ",
            data=rhs)

        # If the roots are different, delete all LHS and add all RHS.
        lhs_is_dict = isinstance(lhs, CommentedMap)
        lhs_is_list = isinstance(lhs, CommentedSeq)
        lhs_is_set = isinstance(lhs, CommentedSet)
        lhs_is_scalar = not (lhs_is_dict or lhs_is_list or lhs_is_set)
        rhs_is_dict = isinstance(rhs, CommentedMap)
        rhs_is_list = isinstance(rhs, CommentedSeq)
        rhs_is_set = isinstance(rhs, CommentedSet)
        rhs_is_scalar = not (rhs_is_dict or rhs_is_list or rhs_is_set)
        same_types = (
            (lhs_is_dict and rhs_is_dict)
            or (lhs_is_list and rhs_is_list)
            or (lhs_is_set and rhs_is_set)
            or (lhs_is_scalar and rhs_is_scalar)
        )
        if same_types:
            if lhs_is_dict:
                self._diff_dicts(path, lhs, rhs)
            elif lhs_is_list:
                self._diff_lists(path, lhs, rhs, **kwargs)
            elif lhs_is_set:
                self._diff_sets(path, lhs, rhs)
            else:
                self._diff_scalars(path, lhs, rhs, **kwargs)
        else:
            self._purge_document(path, lhs)
            self._add_everything(path, rhs)

    @classmethod
    def synchronize_lists_by_value(
        cls, lhs: CommentedSeq, rhs: CommentedSeq
    ) -> List[Tuple[
        Optional[int], Optional[Any], Optional[int], Optional[Any]
    ]]:
        """
        Synchronize two lists by value.

        Parameters:
        1. lhs (Any) The left-hand-side (original) document
        2. rhs (Any) The right-hand-side (altered) document

        Returns:  (List[Tuple[
               Optional[int], Optional[Any], Optional[int], Optional[Any]
            ]]) List with LHS and RHS elements at identical elements
        """
        # Build a parallel index array to track the original RHS element
        # indexes of any surviving elements.
        rhs_reduced = []
        for original_idx, val in enumerate(rhs):
            rhs_reduced.append((original_idx, val))

        syn_pairs: List[Tuple[
            Optional[int], Optional[Any], Optional[int], Optional[Any]
        ]] = []
        for lhs_idx, lhs_ele in enumerate(lhs):
            del_index = -1
            for reduced_idx, rhs_pair in enumerate(rhs_reduced):
                (_, rhs_ele) = rhs_pair
                if rhs_ele == lhs_ele:
                    del_index = reduced_idx
                    break

            if del_index > -1:
                (rhs_original_idx, rhs_ele) = rhs_reduced.pop(del_index)
                syn_pairs.append((lhs_idx, lhs_ele, rhs_original_idx, rhs_ele))
            else:
                syn_pairs.append((lhs_idx, lhs_ele, None, None))

        for rhs_pair in rhs_reduced:
            (rhs_original_idx, rhs_ele) = rhs_pair
            syn_pairs.append((None, None, rhs_original_idx, rhs_ele))

        return syn_pairs

    #pylint: disable=too-many-locals
    def synchronize_lods_by_key(
        self, path: YAMLPath, lhs: CommentedSeq, rhs: CommentedSeq
    ) -> List[Tuple[
        Optional[int], Optional[Any], Optional[int], Optional[Any]
    ]]:
        """
        Synchronize two lists-of-dictionaries by identity key.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document

        Returns: (List[Tuple[
                Optional[int], Optional[Any], Optional[int], Optional[Any]
            ]]) List of identical LHS and RHS elements in the same element
            positions
        """
        # Build a parallel index array to track the original RHS element
        # indexes of any surviving elements.
        key_attr: str = ""
        if len(rhs) > 0 and isinstance(rhs[0], CommentedMap):
            (key_attr, _) = self.config.aoh_diff_key(
                NodeCoords(rhs[0], rhs, 0))
            self.logger.debug(
                "Differ::synchronize_lods_by_key:  RHS AoH yielded key_attr:"
                "  {}.".format(key_attr))

        rhs_reduced = []
        for original_idx, val in enumerate(rhs):
            rhs_reduced.append((original_idx, val))

        syn_pairs: List[Tuple[
            Optional[int], Optional[Any], Optional[int], Optional[Any]
        ]] = []
        for lhs_idx, lhs_ele in enumerate(lhs):
            if not key_attr in lhs_ele:
                # Impossible to match this LHS record to any RHS record
                self.logger.debug(
                    "LHS record has no identity key, {}, for record at {}:"
                    .format(key_attr, path),
                    data=lhs_ele,
                    prefix="Differ::synchronize_lods_by_key:  ")
                syn_pairs.append((lhs_idx, lhs_ele, None, None))
                continue

            del_index = -1
            for reduced_idx, rhs_pair in enumerate(rhs_reduced):
                (rhs_idx, rhs_ele) = rhs_pair

                # Check for a custom identity key assignment for this record
                use_key = key_attr
                (alt_key, is_user_key) = self.config.aoh_diff_key(
                    NodeCoords(rhs_ele, rhs, rhs_idx))

                if is_user_key and alt_key:
                    use_key = alt_key
                    self.logger.debug(
                        "Using alternate key, {}, for record at {}[{}]."
                        .format(use_key, path, rhs_idx),
                        data=rhs_ele,
                        prefix="Differ::synchronize_lods_by_key:  ")
                else:
                    self.logger.debug(
                        "Using inferred key, {}, for record at {}[{}]."
                        .format(use_key, path, rhs_idx),
                        data=rhs_ele,
                        prefix="Differ::synchronize_lods_by_key:  ")

                if not use_key in rhs_ele:
                    # Impossible to match this RHS record to any LHS record
                    continue

                if rhs_ele[use_key] == lhs_ele[use_key]:
                    del_index = reduced_idx
                    break

            if del_index > -1:
                (rhs_original_idx, rhs_ele) = rhs_reduced.pop(del_index)
                syn_pairs.append((lhs_idx, lhs_ele, rhs_original_idx, rhs_ele))
            else:
                syn_pairs.append((lhs_idx, lhs_ele, None, None))

        for rhs_pair in rhs_reduced:
            (rhs_original_idx, rhs_ele) = rhs_pair
            syn_pairs.append((None, None, rhs_original_idx, rhs_ele))

        return syn_pairs

    @classmethod
    def _get_key_indicies(
        cls, data: Union[CommentedMap, CommentedSet]
    ) -> Dict[Any, int]:
        """
        Get a dictionary mapping of keys to relative positions.

        Parameters:
        1. path (YAMLPath) YAML Path to the document element under evaluation
        2. lhs (Any) The left-hand-side (original) document
        3. rhs (Any) The right-hand-side (altered) document

        Returns:  (Dict[Any, int]) Dictionary indicating the element position
            of each key
        """
        key_map = {}
        if isinstance(data, CommentedMap):
            for idx, key in enumerate(data.keys()):
                key_map[key] = idx
        elif isinstance(data, CommentedSet):
            for idx, key in enumerate(data):
                key_map[key] = idx
        return key_map



import pytest

from types import SimpleNamespace

from ruamel.yaml.comments import CommentedMap, CommentedSeq, CommentedSet, TaggedScalar
from ruamel.yaml.scalarstring import PlainScalarString, FoldedScalarString
from ruamel.yaml import version_info as ryversion
if ryversion < (0, 17, 22):                   # pragma: no cover
    from yamlpath.patches.timestamp import (
        AnchoredTimeStamp,
        AnchoredDate,
    )  # type: ignore
else:                                         # pragma: no cover
    # Temporarily fool MYPY into resolving the future-case imports
    from ruamel.yaml.timestamp import TimeStamp as AnchoredTimeStamp
    AnchoredDate = AnchoredTimeStamp
    #from ruamel.yaml.timestamp import AnchoredTimeStamp
    # From whence shall come AnchoredDate?

from yamlpath.enums import PathSegmentTypes
from yamlpath.wrappers import NodeCoords, ConsolePrinter
from yamlpath import YAMLPath

class Test_wrappers_ConsolePrinter():
    def test_info_noisy(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=False, debug=False)
        logger = ConsolePrinter(args)
        logger.info("Test")
        console = capsys.readouterr()
        assert console.out == "Test\n"

    def test_info_quiet(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=True, debug=False)
        logger = ConsolePrinter(args)
        logger.info("Test")
        console = capsys.readouterr()
        assert not console.out

    def test_verbose_off(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=False, debug=False)
        logger = ConsolePrinter(args)
        logger.verbose("Test")
        console = capsys.readouterr()
        assert not console.out

    def test_verbose_noisy(self, capsys):
        args = SimpleNamespace(verbose=True, quiet=False, debug=False)
        logger = ConsolePrinter(args)
        logger.verbose("Test")
        console = capsys.readouterr()
        assert console.out == "Test\n"

    def test_verbose_quiet(self, capsys):
        args = SimpleNamespace(verbose=True, quiet=True, debug=False)
        logger = ConsolePrinter(args)
        logger.verbose("Test")
        console = capsys.readouterr()
        assert not console.out

    def test_debug_off(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=False, debug=False)
        logger = ConsolePrinter(args)
        logger.debug("Test")
        console = capsys.readouterr()
        assert not console.out

    def test_debug_noisy(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=False, debug=True)
        logger = ConsolePrinter(args)
        anchoredkey = PlainScalarString("TestKey", anchor="KeyAnchor")
        anchoredval = PlainScalarString("TestVal", anchor="Anchor")
        foldedstr = "123456789 123456789 123456789"
        foldedstrfolds = [10, 20]
        foldedval = FoldedScalarString(foldedstr)
        foldedval.fold_pos = foldedstrfolds

        logger.debug(anchoredval)
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  (&Anchor)TestVal",
        ]) + "\n" == console.out

        logger.debug(["test", anchoredval])
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  [0]test<class 'str'>",
            "DEBUG:  [1](&Anchor)TestVal<class 'ruamel.yaml.scalarstring.PlainScalarString'>",
        ]) + "\n" == console.out

        logger.debug({"date": AnchoredDate(2022, 9, 23), "timestamp": AnchoredTimeStamp(2022, 9, 25, 1, 2, 3, 40000)})
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  [date]2022-09-23<class 'yamlpath.patches.timestamp.AnchoredDate'>",
            "DEBUG:  [timestamp]2022-09-25T01:02:03.040000<class 'yamlpath.patches.timestamp.AnchoredTimeStamp'>",
        ]) + "\n" == console.out

        logger.debug({"ichi": 1, anchoredkey: anchoredval})
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  [ichi]1<class 'int'>",
            "DEBUG:  [TestKey](&KeyAnchor,&Anchor)TestVal<class 'ruamel.yaml.scalarstring.PlainScalarString'>",
        ]) + "\n" == console.out

        logger.debug({"ichi": 1, anchoredkey: "non-anchored value"})
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  [ichi]1<class 'int'>",
            "DEBUG:  [TestKey](&KeyAnchor,_)non-anchored value<class 'str'>",
        ]) + "\n" == console.out

        logger.debug({"ichi": 1, "non-anchored-key": anchoredval})
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  [ichi]1<class 'int'>",
            "DEBUG:  [non-anchored-key](_,&Anchor)TestVal<class 'ruamel.yaml.scalarstring.PlainScalarString'>",
        ]) + "\n" == console.out

        tagged_value = "value"
        tagged_value_node = TaggedScalar(tagged_value, tag="!tag")
        tagged_sequence = CommentedSeq(["a", "b"])
        tagged_sequence.yaml_set_tag("!raz")
        selfref_value = "self_referring"
        selfref_value_node = TaggedScalar(selfref_value, tag="!self_referring")
        logger.debug(
            "test_wrappers_consoleprinter:",
            prefix="test_debug_noisy:  ",
            header="--- HEADER ---",
            footer="=== FOOTER ===",
            data_header="+++ DATA HEADER +++",
            data_footer="::: DATA FOOTER :::",
            data=CommentedMap({
                "key": "value",
                "tagged": tagged_value_node,
                tagged_value_node: "untagged value",
                selfref_value_node: selfref_value_node,
                "array": ["ichi", "ni", "san"],
                "tagged_array": tagged_sequence,
                "aoh": [{"id": 1},{"id": 2},{"id": 3}],
                "aoa": [[True, True], [True, False], [False, True]],
                "dod": {"a": {"b": {"c": "d"}}},
                "set": CommentedSet(["one", "two"]),
            })
        )
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  test_debug_noisy:  --- HEADER ---",
            "DEBUG:  test_debug_noisy:  test_wrappers_consoleprinter:",
            "DEBUG:  test_debug_noisy:  +++ DATA HEADER +++",
            "DEBUG:  test_debug_noisy:  [key]value<class 'str'>",
            "DEBUG:  test_debug_noisy:  [tagged]<_,!tag>value<class 'ruamel.yaml.comments.TaggedScalar'>(<class 'str'>)",
            "DEBUG:  test_debug_noisy:  [value]<!tag,_>untagged value<class 'str'>",
            "DEBUG:  test_debug_noisy:  [self_referring]<!self_referring,!self_referring>self_referring<class 'ruamel.yaml.comments.TaggedScalar'>(<class 'str'>)",
            "DEBUG:  test_debug_noisy:  [array][0]ichi<class 'str'>",
            "DEBUG:  test_debug_noisy:  [array][1]ni<class 'str'>",
            "DEBUG:  test_debug_noisy:  [array][2]san<class 'str'>",
            "DEBUG:  test_debug_noisy:  [tagged_array]<_,!raz>[0]a<class 'str'>",
            "DEBUG:  test_debug_noisy:  [tagged_array]<_,!raz>[1]b<class 'str'>",
            "DEBUG:  test_debug_noisy:  [aoh][0][id]1<class 'int'>",
            "DEBUG:  test_debug_noisy:  [aoh][1][id]2<class 'int'>",
            "DEBUG:  test_debug_noisy:  [aoh][2][id]3<class 'int'>",
            "DEBUG:  test_debug_noisy:  [aoa][0][0]True<class 'bool'>",
            "DEBUG:  test_debug_noisy:  [aoa][0][1]True<class 'bool'>",
            "DEBUG:  test_debug_noisy:  [aoa][1][0]True<class 'bool'>",
            "DEBUG:  test_debug_noisy:  [aoa][1][1]False<class 'bool'>",
            "DEBUG:  test_debug_noisy:  [aoa][2][0]False<class 'bool'>",
            "DEBUG:  test_debug_noisy:  [aoa][2][1]True<class 'bool'>",
            "DEBUG:  test_debug_noisy:  [dod][a][b][c]d<class 'str'>",
            "DEBUG:  test_debug_noisy:  [set]{one}<class 'str'>",
            "DEBUG:  test_debug_noisy:  [set]{two}<class 'str'>",
            "DEBUG:  test_debug_noisy:  ::: DATA FOOTER :::",
            "DEBUG:  test_debug_noisy:  === FOOTER ===",
        ]) + "\n" == console.out

        logger.debug(tagged_value_node)
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  <!tag>value<class 'ruamel.yaml.comments.TaggedScalar'>(<class 'str'>)",
        ])

        logger.debug(tagged_sequence)
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  [tagged_array]<!raz>[0]a<class 'str'>",
            "DEBUG:  [tagged_array]<!raz>[1]b<class 'str'>",
        ])

        nc = NodeCoords(
            "value",
            dict(key="value"),
            "key",
            YAMLPath("doc_root.key"),
            [   (dict(doc_root=dict(key="value")), "doc_root"),
                (dict(key="value"), "key")],
            (PathSegmentTypes.KEY, "key")
        )
        logger.debug(
            "A node coordinate:", prefix="test_debug_noisy:  ", data=nc)
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  test_debug_noisy:  A node coordinate:",
            "DEBUG:  test_debug_noisy:  (path)doc_root.key",
            "DEBUG:  test_debug_noisy:  (segment)[0]PathSegmentTypes.KEY<enum 'PathSegmentTypes'>",
            "DEBUG:  test_debug_noisy:  (segment)[1]key<class 'str'>",
            "DEBUG:  test_debug_noisy:  (node)value",
            "DEBUG:  test_debug_noisy:  (parent)[key]value<class 'str'>",
            "DEBUG:  test_debug_noisy:  (parentref)key",
            "DEBUG:  test_debug_noisy:  (ancestry)[0][0][doc_root][key]value<class 'str'>",
            "DEBUG:  test_debug_noisy:  (ancestry)[0][1]doc_root<class 'str'>",
            "DEBUG:  test_debug_noisy:  (ancestry)[1][0][key]value<class 'str'>",
            "DEBUG:  test_debug_noisy:  (ancestry)[1][1]key<class 'str'>",
        ]) + "\n" == console.out

        logger.debug(foldedval)
        console = capsys.readouterr()
        assert "\n".join([
            "DEBUG:  {}<class 'ruamel.yaml.scalarstring.FoldedScalarString'>,folded@{}".format(foldedstr, foldedstrfolds)
        ])

    def test_debug_quiet(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=True, debug=True)
        logger = ConsolePrinter(args)
        logger.debug("Test")
        console = capsys.readouterr()
        assert not console.out

    def test_warning_noisy(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=False, debug=False)
        logger = ConsolePrinter(args)
        logger.warning("Test")
        console = capsys.readouterr()
        assert console.out == "WARNING:  Test\n"

    def test_warning_quiet(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=True, debug=False)
        logger = ConsolePrinter(args)
        logger.warning("Test")
        console = capsys.readouterr()
        assert not console.out

    def test_error_noisy_nonexit(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=False, debug=False)
        logger = ConsolePrinter(args)
        logger.error("Test")
        console = capsys.readouterr()
        assert console.err == "ERROR:  Test\n"

    def test_error_quiet_nonexit(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=True, debug=False)
        logger = ConsolePrinter(args)
        logger.error("Test")
        console = capsys.readouterr()
        assert console.err == "ERROR:  Test\n"

    def test_error_noisy_exit(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=False, debug=False)
        logger = ConsolePrinter(args)
        with pytest.raises(SystemExit):
            logger.error("Test", 27)
            console = capsys.readouterr()
            assert console.err == "ERROR:  Test\n"

    def test_error_quiet_exit(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=True, debug=False)
        logger = ConsolePrinter(args)
        with pytest.raises(SystemExit):
            logger.error("Test", 27)
            console = capsys.readouterr()
            assert console.err == "ERROR:  Test\n"

    def test_critical_noisy(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=False, debug=False)
        logger = ConsolePrinter(args)
        with pytest.raises(SystemExit):
            logger.critical("Test")
            console = capsys.readouterr()
            assert console.err == "CRITICAL:  Test\n"

    def test_critical_quiet(self, capsys):
        args = SimpleNamespace(verbose=False, quiet=True, debug=False)
        logger = ConsolePrinter(args)
        with pytest.raises(SystemExit):
            logger.critical("Test")
            console = capsys.readouterr()
            assert console.err == "CRITICAL:  Test\n"



"""
Enable users to rotate EYAML encryption keys used for pre-encrypted files.

Rotates the encryption keys used for all EYAML values within a set of YAML
files, decrypting with old keys and re-encrypting using replacement keys.

Copyright 2018, 2019 William W. Kimball, Jr. MBA MSIS
"""
import sys
import argparse
from shutil import copy2
from os import remove, access, R_OK
from os.path import isfile, exists

from ruamel.yaml.scalarstring import FoldedScalarString

from yamlpath import __version__ as YAMLPATH_VERSION
from yamlpath.common import Anchors, Parsers
from yamlpath.eyaml.exceptions import EYAMLCommandException
from yamlpath.eyaml.enums import EYAMLOutputFormats
from yamlpath.eyaml import EYAMLProcessor
from yamlpath.wrappers import ConsolePrinter

def processcli():
    """Process command-line arguments."""
    parser = argparse.ArgumentParser(
        description=(
            "Rotates the encryption keys used for all EYAML values"
            " within a set of YAML files, decrypting with old keys and"
            " re-encrypting using replacement keys."),
        epilog=(
            "Any YAML_FILEs lacking EYAML values will not be modified (or"
            " backed up, even when -b/--backup is specified).  For more"
            " information about YAML Paths, please visit"
            " https://github.com/wwkimball/yamlpath/wiki.  To report issues"
            " with this tool or to request enhancements, please visit"
            " https://github.com/wwkimball/yamlpath/issues.")
    )
    parser.add_argument("-V", "--version", action="version",
                        version="%(prog)s " + YAMLPATH_VERSION)

    noise_group = parser.add_mutually_exclusive_group()
    noise_group.add_argument("-d", "--debug", action="store_true",
                             help="output debugging details")
    noise_group.add_argument("-v", "--verbose", action="store_true",
                             help="increase output verbosity")
    noise_group.add_argument("-q", "--quiet", action="store_true",
                             help="suppress all output except errors")

    parser.add_argument("-b", "--backup", action="store_true",
                        help="save a backup of each modified YAML_FILE with an"
                        + " extra .bak file-extension")
    parser.add_argument("-x", "--eyaml", default="eyaml",
                        help="the eyaml binary to use when it isn't on the"
                        + " PATH")

    key_group = parser.add_argument_group(
        "EYAML_KEYS", "All key arguments are required"
    )
    key_group.add_argument("-r", "--newprivatekey", required=True,
                           help="the new EYAML private key")
    key_group.add_argument("-u", "--newpublickey", required=True,
                           help="the new EYAML public key")
    key_group.add_argument("-i", "--oldprivatekey", required=True,
                           help="the old EYAML private key")
    key_group.add_argument("-c", "--oldpublickey", required=True,
                           help="the old EYAML public key")

    parser.add_argument("yaml_files", metavar="YAML_FILE", nargs="+",
                        help="one or more YAML files containing EYAML values")
    return parser.parse_args()

def validateargs(args, log):
    """Validate command-line arguments."""
    has_errors = False

    # Enforce sanity
    # * The new and old EYAML keys must be different
    if ((args.newprivatekey == args.oldprivatekey)
            or (args.newpublickey == args.oldpublickey)):
        has_errors = True
        log.error("The new and old EYAML keys must be different.")

    # * All EYAML certs must exist and be readable to the present user
    for check_file in [args.newprivatekey,
                       args.newpublickey,
                       args.oldprivatekey,
                       args.oldpublickey]:
        if not (isfile(check_file) and access(check_file, R_OK)):
            has_errors = True
            log.error(
                "EYAML key is not a readable file:  " + check_file
            )

    if has_errors:
        sys.exit(1)

# pylint: disable=locally-disabled,too-many-locals,too-many-branches,too-many-statements
def main():
    """Main code."""
    # Process any command-line arguments
    args = processcli()
    log = ConsolePrinter(args)
    validateargs(args, log)
    processor = EYAMLProcessor(log, None, binary=args.eyaml)

    # Prep the YAML parser
    yaml = Parsers.get_yaml_editor()

    # Process the input file(s)
    in_file_count = len(args.yaml_files)
    exit_state = 0
    for yaml_file in args.yaml_files:
        file_changed = False
        backup_file = yaml_file + ".bak"
        seen_anchors = []

        # Each YAML_FILE must actually be a file
        if not isfile(yaml_file):
            log.error("Not a file:  {}".format(yaml_file))
            exit_state = 2
            continue

        # Don't bother with the file change update when there's only one input
        # file.
        if in_file_count > 1:
            log.info("Processing {}...".format(yaml_file))

        # Try to open the file
        (yaml_data, doc_loaded) = Parsers.get_yaml_data(yaml, log, yaml_file)
        if not doc_loaded:
            # An error message has already been logged
            exit_state = 3
            continue

        # Process all EYAML values
        processor.data = yaml_data
        for yaml_path in processor.find_eyaml_paths():
            # Use ::get_nodes() instead of ::get_eyaml_values() here in order
            # to ignore values that have already been rotated via their
            # Anchors.
            for node_coordinate in processor.get_nodes(
                yaml_path, mustexist=True
            ):
                # Ignore values which are Aliases for those already decrypted
                node = node_coordinate.node
                anchor_name = Anchors.get_node_anchor(node)
                if anchor_name is not None:
                    if anchor_name in seen_anchors:
                        continue

                    seen_anchors.append(anchor_name)

                log.verbose("Decrypting value(s) at {}.".format(yaml_path))
                processor.publickey = args.oldpublickey
                processor.privatekey = args.oldprivatekey

                try:
                    txtval = processor.decrypt_eyaml(node)
                except EYAMLCommandException as ex:
                    log.error(ex)
                    exit_state = 3
                    continue

                # Prefer block (folded) values unless the original YAML value
                # was already a massivly long (string) line.
                output = EYAMLOutputFormats.BLOCK
                if not isinstance(node, FoldedScalarString):
                    output = EYAMLOutputFormats.STRING

                # Re-encrypt the value with new EYAML keys
                processor.publickey = args.newpublickey
                processor.privatekey = args.newprivatekey

                try:
                    processor.set_eyaml_value(yaml_path, txtval, output=output)
                except EYAMLCommandException as ex:
                    log.error(ex)
                    exit_state = 3
                    continue

                file_changed = True

        # Save the changes
        if file_changed:
            if args.backup:
                log.verbose("Saving a backup of {} to {}."
                            .format(yaml_file, backup_file))
                if exists(backup_file):
                    remove(backup_file)
                copy2(yaml_file, backup_file)

            log.verbose("Writing changed data to {}.".format(yaml_file))
            with open(yaml_file, 'w', encoding='utf-8') as yaml_dump:
                yaml.dump(yaml_data, yaml_dump)

    sys.exit(exit_state)

if __name__ == "__main__":
    main()  # pragma: no cover



import pytest

import ruamel.yaml as ry

from yamlpath.enums import PathSearchKeywords
from yamlpath.path import SearchKeywordTerms
from yamlpath.common import KeywordSearches
from yamlpath.exceptions import YAMLPathException
from yamlpath import YAMLPath

class Test_common_keywordsearches():
    """Tests for the KeywordSearches helper class."""

    ###
    # search_matches
    ###
    def test_unknown_search_keyword(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.search_matches(
                SearchKeywordTerms(False, None, ""),
                {},
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("Unsupported search keyword")


    ###
    # has_child
    ###
    def test_has_child_invalid_param_count(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.search_matches(
                SearchKeywordTerms(False, PathSearchKeywords.HAS_CHILD, []),
                {},
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("Invalid parameter count to ")

    def test_has_child_invalid_node(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.has_child(
                "abc: xyz",
                False,
                ["wwk"],
                YAMLPath("")
            ))
        assert -1 < str(ex.value).find("has no child nodes")


    ###
    # name
    ###
    def test_name_invalid_param_count(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.name(
                False,
                ["1", "2"],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("Invalid parameter count to ")

    def test_name_invalid_inversion(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.name(
                True,
                [],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("Inversion is meaningless to ")


    ###
    # max
    ###
    def test_max_invalid_param_count(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.max(
                {},
                False,
                ["1", "2"],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("Invalid parameter count to ")

    def test_max_missing_aoh_param(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.max(
                [{'a': 1},{'a': 2}],
                False,
                [],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("when evaluating an Array-of-Hashes")

    def test_max_missing_hash_param(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.max(
                {'a': {'b': 1}, 'c': {'d': 2}},
                False,
                [],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("when comparing Hash/map/dict children")

    def test_max_invalid_array_param(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.max(
                [1, 2, 3],
                False,
                ['3'],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("when comparing Array/sequence/list elements to one another")

    def test_max_incorrect_node(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.max(
                {'b': 2},
                False,
                ['b'],
                YAMLPath("/*[max(b)]")
            ))
        assert -1 < str(ex.value).find("operates against collections of data")


    ###
    # min
    ###
    def test_min_invalid_param_count(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.min(
                {},
                False,
                ["1", "2"],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("Invalid parameter count to ")

    def test_min_missing_aoh_param(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.min(
                [{'a': 1},{'a': 2}],
                False,
                [],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("when evaluating an Array-of-Hashes")

    def test_min_missing_hash_param(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.min(
                {'a': {'b': 1}, 'c': {'d': 2}},
                False,
                [],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("when comparing Hash/map/dict children")

    def test_min_invalid_array_param(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.min(
                [1, 2, 3],
                False,
                ['3'],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("when comparing Array/sequence/list elements to one another")

    def test_min_incorrect_node(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.min(
                {'b': 2},
                False,
                ['b'],
                YAMLPath("/*[max(b)]")
            ))
        assert -1 < str(ex.value).find("operates against collections of data")


    ###
    # parent
    ###
    def test_parent_invalid_param_count(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.parent(
                {},
                False,
                ["1", "2"],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("Invalid parameter count to ")

    def test_parent_invalid_inversion(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.parent(
                {},
                True,
                [],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("Inversion is meaningless to ")

    def test_parent_invalid_parameter(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.parent(
                {},
                False,
                ["abc"],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("Invalid parameter passed to ")

    def test_parent_invalid_step_count(self):
        with pytest.raises(YAMLPathException) as ex:
            nodes = list(KeywordSearches.parent(
                {},
                False,
                ["5"],
                YAMLPath("/")
            ))
        assert -1 < str(ex.value).find("higher than the document root")



import pytest
from datetime import datetime, date

from ruamel.yaml.scalarstring import (
    PlainScalarString,
    DoubleQuotedScalarString,
    SingleQuotedScalarString,
    FoldedScalarString,
    LiteralScalarString,
)
from ruamel.yaml.scalarbool import ScalarBoolean
from ruamel.yaml.scalarfloat import ScalarFloat
from ruamel.yaml.scalarint import ScalarInt
from ruamel.yaml import version_info as ryversion
if ryversion < (0, 17, 22):                   # pragma: no cover
    from yamlpath.patches.timestamp import (
        AnchoredTimeStamp,
    )  # type: ignore
else:
    # Temporarily fool MYPY into resolving the future-case imports
    from ruamel.yaml.timestamp import TimeStamp as AnchoredTimeStamp
    #from ruamel.yaml.timestamp import AnchoredTimeStamp

from yamlpath.enums import YAMLValueFormats


class Test_enums_YAMLValueFormats():
	"""Tests for the YAMLValueFormats enumeration."""
	def test_get_names(self):
		assert YAMLValueFormats.get_names() == [
			"BARE",
			"BOOLEAN",
			"DATE",
			"DEFAULT",
			"DQUOTE",
			"FLOAT",
			"FOLDED",
			"INT",
			"LITERAL",
			"SQUOTE",
			"TIMESTAMP",
		]

	@pytest.mark.parametrize("input,output", [
		("BARE", YAMLValueFormats.BARE),
		("BOOLEAN", YAMLValueFormats.BOOLEAN),
		("DATE", YAMLValueFormats.DATE),
		("DEFAULT", YAMLValueFormats.DEFAULT),
		("DQUOTE", YAMLValueFormats.DQUOTE),
		("FLOAT", YAMLValueFormats.FLOAT),
		("FOLDED", YAMLValueFormats.FOLDED),
		("INT", YAMLValueFormats.INT),
		("LITERAL", YAMLValueFormats.LITERAL),
		("SQUOTE", YAMLValueFormats.SQUOTE),
		("TIMESTAMP", YAMLValueFormats.TIMESTAMP),
	])
	def test_from_str(self, input, output):
		assert output == YAMLValueFormats.from_str(input)

	def test_from_str_nameerror(self):
		with pytest.raises(NameError):
			YAMLValueFormats.from_str("NO SUCH NAME")

	@pytest.mark.parametrize("input,output", [
		(FoldedScalarString(""), YAMLValueFormats.FOLDED),
		(LiteralScalarString(""), YAMLValueFormats.LITERAL),
		(date(2022, 9, 24), YAMLValueFormats.DATE),
		(DoubleQuotedScalarString(''), YAMLValueFormats.DQUOTE),
		(SingleQuotedScalarString(""), YAMLValueFormats.SQUOTE),
		(PlainScalarString(""), YAMLValueFormats.BARE),
		(ScalarBoolean(False), YAMLValueFormats.BOOLEAN),
		(ScalarFloat(1.01), YAMLValueFormats.FLOAT),
		(ScalarInt(10), YAMLValueFormats.INT),
		(AnchoredTimeStamp(2022, 9, 24, 7, 42, 38), YAMLValueFormats.TIMESTAMP),
		(None, YAMLValueFormats.DEFAULT),
	])
	def test_from_node(self, input, output):
		assert output == YAMLValueFormats.from_node(input)



"""
Implement SearchKeywordTerms.

Copyright 2021 William W. Kimball, Jr. MBA MSIS
"""
from typing import List

from yamlpath.enums import PathSearchKeywords


class SearchKeywordTerms:
    """YAML path Search Keyword segment terms."""

    def __init__(
        self, inverted: bool, keyword: PathSearchKeywords, parameters: str
    ) -> None:
        """
        Instantiate a Keyword Search Term segment.

        Parameters:
        1. inverted (bool) true = invert the search operation; false, otherwise
        2. keyword (PathSearchKeywords) the search keyword
        3. parameters (str) the parameters to the keyword-named operation
        """
        self._inverted: bool = inverted
        self._keyword: PathSearchKeywords = keyword
        self._parameters: str = parameters
        self._lparameters: List[str] = []
        self._parameters_parsed: bool = False

    def __str__(self) -> str:
        """Get a String representation of this Keyword Search Term."""
        return (
            "["
            + ("!" if self._inverted else "")
            + str(self._keyword)
            + "("
            + self._parameters
            + ")]"
        )

    @property
    def inverted(self) -> bool:
        """
        Access the inversion flag for this Keyword Search.

        This indicates whether the search logic is to be inverted.
        """
        return self._inverted

    @property
    def keyword(self) -> PathSearchKeywords:
        """
        Access the search keyword.

        This indicates what kind of search logic is to be performed.
        """
        return self._keyword

    @property
    # pylint: disable=locally-disabled,too-many-branches
    def parameters(self) -> List[str]:
        """Accessor for the parameters being fed to the search operation."""
        if self._parameters_parsed:
            return self._lparameters

        if self._parameters is None:
            self._parameters_parsed = True
            self._lparameters = []
            return self._lparameters

        param: str = ""
        params: List[str] = []
        escape_next: bool = False
        demarc_stack: List[str] = []
        demarc_count: int = 0

        # pylint: disable=locally-disabled,too-many-nested-blocks
        for char in self._parameters:
            demarc_count = len(demarc_stack)

            if escape_next:
                # Pass-through; capture this escaped character
                escape_next = False

            elif char == "\\":
                escape_next = True
                continue

            elif (
                    char == " "
                    and (demarc_count < 1)
            ):
                # Ignore unescaped, non-demarcated whitespace
                continue

            elif char in ['"', "'"]:
                # Found a string demarcation mark
                if demarc_count > 0:
                    # Already appending to an ongoing demarcated value
                    if char == demarc_stack[-1]:
                        # Close a matching pair
                        demarc_stack.pop()
                        demarc_count -= 1

                        if demarc_count < 1:
                            # Final close; seek the next delimiter
                            continue

                    else:
                        # Embed a nested, demarcated component
                        demarc_stack.append(char)
                        demarc_count += 1
                else:
                    # Fresh demarcated value
                    demarc_stack.append(char)
                    demarc_count += 1
                    continue

            elif demarc_count < 1 and char == ",":
                params.append(param)
                param = ""
                continue

            param = param + char

        # Check for mismatched demarcations
        if demarc_count > 0:
            raise ValueError(
                "Keyword search parameters contain one or more unmatched"
                " demarcation symbol(s): {}".format(" ".join(demarc_stack)))

        # Add the last parameter, if there is one
        if param:
            params.append(param)

        self._lparameters = params
        self._parameters_parsed = True
        return self._lparameters



"""Make all the YAML Path Merger enumerations available."""
from .anchorconflictresolutions import AnchorConflictResolutions
from .aohmergeopts import AoHMergeOpts
from .arraymergeopts import ArrayMergeOpts
from .hashmergeopts import HashMergeOpts
from .multidocmodes import MultiDocModes
from .outputdoctypes import OutputDocTypes
from .setmergeopts import SetMergeOpts



"""
Represents an exception that occurs during an EYAML command execution.

Copyright 2019 William W. Kimball, Jr. MBA MSIS
"""


class EYAMLCommandException(Exception):
    """Exception to raise when an EYAML command executation fails."""



"""
Implements the CollectorOperators enumeration.

Copyright 2019, 2020 William W. Kimball, Jr. MBA MSIS
"""
from enum import Enum, auto
from typing import List


class CollectorOperators(Enum):
    """
    Supported Collector operators.

    These identify how one Collector's results are to be combined with its
    predecessor Collector, if there is one.  Operations include:

    `NONE`
        The Collector's results are not combined with its predecessor.
        Instead, the Collector creates a new result derived from its position
        with the data.

    `ADDITION`
        The Collector's results are concatenated with its immediate predecessor
        Collector's results.  No effort is made to limit the resluting data to
        unique values.

    `SUBTRACTION`
        The Collector's results are removed from its immediate predecessor
        Collector's results.  Only exact matches are removed.
    """

    ADDITION = auto()
    NONE = auto()
    SUBTRACTION = auto()

    def __str__(self) -> str:
        """Get a String representation of an employed value of this enum."""
        operator: str = ''
        if self is CollectorOperators.ADDITION:
            operator = '+'
        elif self is CollectorOperators.SUBTRACTION:
            operator = '-'
        return operator

    @staticmethod
    def get_names() -> List[str]:
        """
        Get all entry names for this enumeration.

        Parameters:  N/A

        Returns:  (List[str]) Upper-case names from this enumeration

        Raises:  N/A
        """
        return [entry.name.upper() for entry in CollectorOperators]

    @staticmethod
    def from_operator(operator: str) -> "CollectorOperators":
        """
        Convert a string value to a value of this enumeration, if valid.

        Parameters:
            1. operator (str) The name to convert

        Returns:  (CollectorOperators) the converted enumeration value

        Raises:
            - `NameError` when name doesn't match any enumeration values.
        """
        if isinstance(operator, CollectorOperators):
            return operator

        check: str = str(operator).upper()

        if check == '+':
            check = "ADDITION"
        elif check == '-':
            check = "SUBTRACTION"

        if check in CollectorOperators.get_names():
            return CollectorOperators[check]
        raise NameError(
            "CollectorOperators has no such item, {}.".format(check))
